%{
#include <stdio.h>
#include <stdlib.h>
#include "messages.h"
#include "sv_circuit_imp.h"
#include "properties.h"

//#define DEBUG_YACC 1
#ifdef DEBUG_YACC
  #define DEBUG_YACC_MESSAGE(msg) fprintf(stderr, msg)
#else
  #define DEBUG_YACC_MESSAGE(msg) // msg
#endif

extern int properties_lex();

void properties_error(CircuitImp* circuit, const char* file_name, const char* message)
{
  Error("%s in line %d in file <%s>", message, properties_lineno, file_name);
}

int implementation_count;
int implementation_index;
%}

%parse-param {CircuitImp* circuit}
%parse-param {const char* file_name}

%start source_text

%union
{
  char* string;
  int inumber;
  Implementation* implementation;
  ImplementationArray* implementations;
}

%type <implementation> name
%type <implementations> implementations
%type <implementations> implementations_ordered
%type <implementations> implementation_ordered
%type <implementations> implementations_named
%type <implementations> implementation_named
%type <implementations> variations_opt
%type <implementations> variations
%type <implementation> variation
%type <string> variables
%type <string> variable
%type <inumber> value
%token <string> YYID
%token <inumber> YYUNDEFINED
%token <inumber> YYINUMBER

%%

source_text
  : names_opt properties_opt
    {
      DEBUG_YACC_MESSAGE("source_text : names_opt properties_opt\n");
    }
;

names_opt
  :
    {
      DEBUG_YACC_MESSAGE("names_opt : \n");
    }
  | '[' names ']'
    {
      DEBUG_YACC_MESSAGE("names_opt : '[' names ']'\n");
    }
;

names
  : name
    {
      DEBUG_YACC_MESSAGE("names : name\n");
      implementation_count = circuit->implementations->count();
      circuit->addImplementation($1);
    }
  | names ',' name
    {
      DEBUG_YACC_MESSAGE("names : names ',' name\n");
      circuit->addImplementation($3);
    }
;

name
  : YYID
    {
      DEBUG_YACC_MESSAGE("name : YYID\n");
      $$ = new Implementation($1, IMPLEMENTATION_TYPE_UNDEFINED);
      delete [] $1;
    }
  | YYID ':' '-'
    {
      DEBUG_YACC_MESSAGE("name : YYID ':' '-'\n");
      $$ = new Implementation($1, IMPLEMENTATION_TYPE_NONE);
      delete [] $1;
    }
  | YYID ':' '+'
    {
      DEBUG_YACC_MESSAGE("name : YYID ':' '+'\n");
      $$ = new Implementation($1, IMPLEMENTATION_TYPE_PARTIAL);
      delete [] $1;
    }
  | YYID ':' '*'
    {
      DEBUG_YACC_MESSAGE("name : YYID ':' '*'\n");
      $$ = new Implementation($1, IMPLEMENTATION_TYPE_FULL);
      delete [] $1;
    }
;

properties_opt
  :
    {
      DEBUG_YACC_MESSAGE("properties_opt : \n");
    }
  | properties
    {
      DEBUG_YACC_MESSAGE("properties_opt : items\n");
    }
;

properties
  : property
    {
      DEBUG_YACC_MESSAGE("properties : property\n");
    }
  | properties property
    {
      DEBUG_YACC_MESSAGE("properties : properties property\n");
    }
;

property
  : YYID implementations
    {
      DEBUG_YACC_MESSAGE("property : ID implementations\n");
      Property* property_cur = circuit->findProperty($1);
      if (property_cur)
        property_cur->addImplementations($2);
      else
      {
        property_cur = new Property($1, $2);
        circuit->addProperty(property_cur);
      }
      delete [] $1;
      delete $2;
    }
;

implementations
  : implementations_ordered
    {
      $$ = $1;
    }
  | implementations_named
    {
      $$ = $1;
    }
;

implementations_ordered
  : implementation_ordered
    {
      DEBUG_YACC_MESSAGE("implementations_ordered : implementation_ordered\n");
      $$ = new ImplementationArray(1);
      implementation_index = implementation_count;
      Implementation* implementation_cur = (*circuit->implementations)[implementation_index];
      for (int variation_index = 0; variation_index < $1->count(); variation_index++)
      {
        Implementation* variation_cur = (*$1)[variation_index];
        char* name_new = name2pattern("%s%s", implementation_cur->getName(), variation_cur->getName());
        variation_cur->setName(name_new);
        delete [] name_new;
        variation_cur->setType(implementation_cur->getType());
        $$->add(variation_cur);
      }
      delete $1;
    }
  | implementations_ordered ';' implementation_ordered
    {
      DEBUG_YACC_MESSAGE("implementations_ordered : implementations_ordered ',' implementation_ordered\n");
      $$ = $1;
      implementation_index++;
      Implementation* implementation_cur = (*circuit->implementations)[implementation_index];
      for (int variation_index = 0; variation_index < $3->count(); variation_index++)
      {
        Implementation* variation_cur = (*$3)[variation_index];
        char* name_new = name2pattern("%s%s", implementation_cur->getName(), variation_cur->getName());
        variation_cur->setName(name_new);
        delete [] name_new;
        variation_cur->setType(implementation_cur->getType());
        $$->add(variation_cur);
      }
      delete $3;
    }
;

implementation_ordered
  : value
    {
      DEBUG_YACC_MESSAGE("implementation_ordered : value\n");
      $$ = new ImplementationArray(1);
      $$->add( new Implementation("", IMPLEMENTATION_TYPE_UNDEFINED, $1) );
    }
  | variations_opt
    {
      DEBUG_YACC_MESSAGE("implementation_ordered : variations_opt\n");
      $$ = $1;
    }
;

implementations_named
  : implementation_named
    {
      DEBUG_YACC_MESSAGE("implementations_named : implementation_named\n");
      $$ = new ImplementationArray(1);
      for (int implementation_index = 0; implementation_index < $1->count(); implementation_index++)
        $$->add( (*$1)[implementation_index] );
      delete $1;
    }
  | implementations_named ';' implementation_named
    {
      DEBUG_YACC_MESSAGE("implementations_named : implementations_named ';' implementation_named\n");
      $$ = $1;
      for (int implementation_index = 0; implementation_index < $3->count(); implementation_index++)
        $$->add( (*$3)[implementation_index] );
      delete $3;
    }
;

implementation_named
  : YYID '=' value
    {
      DEBUG_YACC_MESSAGE("implementation_named : ID '=' value\n");
      $$ = new ImplementationArray(1);
      Implementation* implementation_cur = circuit->findImplementation($1);
      $$->add( new Implementation($1, implementation_cur->getType(), $3) );
      delete [] $1;
    }
  | YYID '=' variations_opt
    {
      DEBUG_YACC_MESSAGE("implementation_ordered : ID '=' variations_opt\n");
      $$ = $3;
      for (int variation_index = 0; variation_index < $$->count(); variation_index++)
      {
        Implementation* variation_cur = (*$$)[variation_index];
        char* name_new = name2pattern("%s%s", $1, variation_cur->getName());
        variation_cur->setName(name_new);
        delete [] name_new;
      }
      delete [] $1;
    }
;
 
variations_opt
  : '(' ')'
    {
      DEBUG_YACC_MESSAGE("variations_opt : '(' ')'\n");
      $$ = new ImplementationArray(0);
    }
  | '(' variations ')'
    {
      DEBUG_YACC_MESSAGE("variations_opt : '(' variations ')'\n");
      $$ = $2;
    }
;

variations
  : variation
    {
      DEBUG_YACC_MESSAGE("variations : variation\n");
      $$ = new ImplementationArray(8);
      $$->add($1);
    }
  | variations ',' variation
    {
      DEBUG_YACC_MESSAGE("variations : variations ',' variation\n");
      $$ = $1;
      $$->add($3);
    }
;

variation
  : variables ':' value
    {
      DEBUG_YACC_MESSAGE("variation : variables ':' value\n");
      char* name_new = name2pattern("(%s)", $1);
      delete [] $1;
      $$ = new Implementation(name_new, IMPLEMENTATION_TYPE_PARTIAL, $3);
      delete [] name_new;
    }
;

variables
  : variable
    {
      DEBUG_YACC_MESSAGE("variables : variable\n");
      $$ = namecopy($1);
      delete [] $1;
    }
  | variables variable
    {
      DEBUG_YACC_MESSAGE("variables : variables variable\n");
      $$ = name2pattern("%s,%s", $1, $2);
      delete [] $1;
      delete [] $2;
    }
;

variable
  : YYID
    {
      DEBUG_YACC_MESSAGE("variable : ID\n");
      $$ = $1;
    }
; 

value
  : YYINUMBER
    {
      DEBUG_YACC_MESSAGE("value : INUMBER\n");
      $$ = $1;
    }
  | YYUNDEFINED
    {
      DEBUG_YACC_MESSAGE("value : INUMBER\n");
      $$ = $1;
    }
;  

%%
