#include "messages.h"
#include "sverilog.h"
#include "sv_circuit.h"
#include "sv_object.h"
#include "sv_module.h"

//---------------------------------------------------------------------------
const char* VERILOG_KEYWORD_MODULE    = "module";
const char* VERILOG_KEYWORD_INPUT     = "input";
const char* VERILOG_KEYWORD_OUTPUT    = "output";
const char* VERILOG_KEYWORD_INOUT     = "inout";
const char* VERILOG_KEYWORD_WIRE      = "wire";
const char* VERILOG_KEYWORD_REG       = "reg";
const char* VERILOG_KEYWORD_UNDEFINED = "undefined";
const char* VERILOG_KEYWORD_PARAMETER = "parameter";
const char* VERILOG_KEYWORD_ENDMODULE = "endmodule";

const int MAX_LINE_LENGTH = 120;

//---------------------------------------------------------------------------
Circuit::Circuit(const char* _name, ModuleArray* _modules, Tag _tag, void* _data) 
: Object(_name, _tag, _data)
{
  module_hash_set = new ModuleHashSet();
  if (_modules)
    modules = new ModuleArray(_modules->count());
  else 
    modules = new ModuleArray(10);
  addModules(_modules);
}

//---------------------------------------------------------------------------
Circuit::Circuit(const Circuit& circuit) 
: Object(circuit)
{
  module_hash_set = new ModuleHashSet();
  modules = new ModuleArray(circuit.modules->count());
  for (int module_index = 0; module_index < circuit.modules->count(); module_index++)
  {
    Module* module_cur = (*circuit.modules)[module_index];
    addModule( new Module(*module_cur) );
  }
}

//---------------------------------------------------------------------------
Circuit::~Circuit()
{
  clean();
  delete module_hash_set; module_hash_set = NULL;
  delete modules; modules = NULL;
}

//---------------------------------------------------------------------------
void Circuit::clean()
{
  modules->clean(true);
}

//---------------------------------------------------------------------------
void Circuit::addModule(Module* module)
{
  if (module_hash_set->find(module) == NULL)
  {
    module->setContainer(this);
    modules->add(module);
    module_hash_set->add(module);
  }
  else
  {
    Warning("module <%s> repeats in the circuit", module->getName());
    delete module;
  }
}

//---------------------------------------------------------------------------
void Circuit::addModules(ModuleArray* _modules)
{
  if (_modules)
  for (int module_index = 0; module_index < _modules->count(); module_index++)
  {
    Module* module_cur = (*_modules)[module_index];
    addModule(module_cur);
  }
}

//---------------------------------------------------------------------------
void Circuit::removeModule(Module* module)
{
  module_hash_set->remove(module);
  modules->remove(module, true);
}

//---------------------------------------------------------------------------
Module* Circuit::findModule(const char* module_name) const
{
  Module* module_mask = new Module(module_name);
  Module* module_cur = module_hash_set->find(module_mask);
  delete module_mask;
  return module_cur;
}

//---------------------------------------------------------------------------
Module* Circuit::setTop(const char* module_name)
{
  // try to find topmost module by name
  top = findModule(module_name);
  if (top == NULL)
    Error("unable to find the topmost module <%s>", module_name);
  return top;
}

//---------------------------------------------------------------------------
Module* Circuit::getTop() const
{
  return top;
}

//---------------------------------------------------------------------------
Module* Circuit::calcTop()
{
  top = NULL;
  ModuleArray* modules_internal = new ModuleArray(1);
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
    if (module_cur->getType() == MODULE_TYPE_COMPOUND)
    {
      for (int instance_index = 0; instance_index < module_cur->instances->count(); instance_index++)
      {
        Instance* instance_cur = (*module_cur->instances)[instance_index];
        Module* module_internal = findModule(instance_cur->getModuleName());
        if (module_internal)
          modules_internal->add(module_internal);
      }
    }
  }
  // if there is at least 1 not internal module
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
    if (module_cur->getType() == MODULE_TYPE_COMPOUND)
    {
      if (modules_internal->locate(module_cur) == -1)
      {
        if (top)
        {
          top = NULL;
          break;
        }
        else
          top = module_cur;
      }
    }
  }
  delete modules_internal; modules_internal = NULL;
  if (top)
    Message("the top module <%s> detected", top->getName());
  else
    Error("unable to detect a top module");
  return top;
}

//---------------------------------------------------------------------------
void Circuit::calcConnections()
{
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
//    if (module_cur->getType() == MODULE_TYPE_COMPOUND)
      module_cur->calcConnections();
  }
}

//---------------------------------------------------------------------------
void Circuit::sortConnections()
{
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
    if (module_cur->getType() == MODULE_TYPE_COMPOUND)
      module_cur->sortConnections();
  }
}

//---------------------------------------------------------------------------
void Circuit::findHierarchy(Module* module, int depth)
{
  if ( !module->testTag(TAG_PROCESSED) )
  if (depth > module->getDepth())
  {
    module->setDepth(depth);
    module->addTag(TAG_PROCESSED);
    if (modules)
    for (int module_index = 0; module_index < modules->count(); module_index++)
    {
      Module* module_cur = (*modules)[module_index];
      if (module_cur->users)
      for (int user_index = 0; user_index < module_cur->users->count(); user_index++)
      {
        Instance* user_cur = (*module_cur->users)[user_index];
        if (user_cur->getContainer() == module)
          findHierarchy(module_cur, depth+1);
      }
    }
    module->removeTag(TAG_PROCESSED);
  }
}

//---------------------------------------------------------------------------
void Circuit::calcHierarchy()
{
  // clear depth and processed flag
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
    module_cur->setDepth(-1);
    module_cur->removeTag(TAG_PROCESSED);
  }
  // calc the modules users
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
    for (int instance_index = 0; instance_index < module_cur->instances->count(); instance_index++)
    {
      Instance* instance_cur = (*module_cur->instances)[instance_index];
      Module* module_prototype = findModule(instance_cur->getModuleName());
      if (module_prototype)
        module_prototype->addUser(instance_cur);
    }
  }
  // calc the depth of the modules starting from the top level modules (which are not instantiated)
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
    if (module_cur->users->count() == 0)
      findHierarchy(module_cur, 0);
  }
  // calc the incidence of instances each module is built of
  for (int depth_cur = getDepth(); depth_cur >= 0; depth_cur--)
  {
    for (int module_index = 0; module_index < modules->count(); module_index++)
    {
      Module* module_cur = (*modules)[module_index];
      if (module_cur->getType() == MODULE_TYPE_COMPOUND)
        if (module_cur->getDepth() == depth_cur)
          module_cur->calcHierarchy();
    }
  }
}

//---------------------------------------------------------------------------
int Circuit::getDepth() const
{
  int depth_max = -1;
  for (int module_index = 0; module_index < modules->count(); module_index++)
  {
    Module* module_cur = (*modules)[module_index];
    int depth_cur = module_cur->getDepth();
    if (depth_cur > depth_max)
      depth_max = depth_cur;
  }
  return depth_max;
}

//---------------------------------------------------------------------------
bool Circuit::isHierarchical() const
{
  return (getDepth() == 0);
}

//---------------------------------------------------------------------------
void Circuit::readVerilog(const char* file_names, ModuleType module_type)
{
  if (file_names)
  {
    int name_index = 0;
    while (char* file_name = names2name(file_names, name_index++))
    {
      if (strcmp(file_name, NAME_STDIO) == 0)
        sverilog_in = stdin;
      else
        if ((sverilog_in = fopen(file_name, "r")) == NULL)
          Error("opening input file <%s>", file_name);
      sverilog_parse(this, file_name, module_type);
      if (sverilog_in != stdin)
        fclose(sverilog_in);
      delete [] file_name;
    }
    calcConnections();
    sortConnections();
  }
}
  
//---------------------------------------------------------------------------
void Circuit::writeVerilogModuleInterface(FILE* file, Module* module) const
{
  fprintf(file, "%s %s", VERILOG_KEYWORD_MODULE, module->getName());
  int line_length = strlen(VERILOG_KEYWORD_MODULE) + 2 + strlen(module->getName()) + 2;
  bool port_first = true;
  for (int port_index = 0; port_index < module->ports->count(); port_index++)
  {
    Port* port_cur = (*module->ports)[port_index];
    if (!port_cur->testTag(TAG_HIDDEN))
    {
      if ( (port_cur->getType() == PORT_TYPE_INPUT)
        || (port_cur->getType() == PORT_TYPE_OUTPUT) 
        || (port_cur->getType() == PORT_TYPE_INOUT) )
      {
        if (port_first)
        {
          fprintf(file, " (");
          port_first = false;
        }
        else
          fprintf(file, ", ");

        if (line_length + strlen(port_cur->getName()) > MAX_LINE_LENGTH)
        {
          line_length = 2;
          fprintf(file, "\n  %s", port_cur->getName());
        }
        else
          fprintf(file, "%s", port_cur->getName());
        line_length += strlen( port_cur->getName()) + 2;
      }
    }
  }
  fprintf(file, ");\n");
}  
//---------------------------------------------------------------------------
void Circuit::writeVerilogModulePorts(FILE* file, Module* module) const
{
  for (int port_index = 0; port_index < module->ports->count(); port_index++)
  {
    Port* port_cur = (*module->ports)[port_index];
    if ( !port_cur->testTag(TAG_HIDDEN) )
    {
      switch (port_cur->getType())
      {
        case PORT_TYPE_INPUT:
          fprintf(file, "  %s", VERILOG_KEYWORD_INPUT);
          break;
        case PORT_TYPE_OUTPUT:
          fprintf(file, "  %s", VERILOG_KEYWORD_OUTPUT);
          break;
        case PORT_TYPE_INOUT:
          fprintf(file, "  %s", VERILOG_KEYWORD_INOUT);
          break;
        case PORT_TYPE_WIRE:
          fprintf(file, "  %s", VERILOG_KEYWORD_WIRE);
          break;
        case PORT_TYPE_REG:
          fprintf(file, "  %s", VERILOG_KEYWORD_REG);
          break;
      }
      Range* range_cur = port_cur->getRange();
      if (range_cur)
        fprintf(file, " [%d:%d]", range_cur->getFrom(), range_cur->getTo());
      fprintf(file, " %s;", port_cur->getName());
      if (port_cur->nets->count() == 1)
        fprintf(file, "  // depth=%d, layer=%d", (*port_cur->nets)[0]->getDepth(), (*port_cur->nets)[0]->getLayer());
      fprintf(file, "\n");
      
    }
  }
}

//---------------------------------------------------------------------------
void Circuit::writeVerilogModuleParameters(FILE* file, Module* module) const
{
  bool parameter_first = true;
  for (int parameter_index = 0; parameter_index < module->parameters->count(); parameter_index++)
  {
    Parameter* parameter_cur = (*module->parameters)[parameter_index];
    if (parameter_first)
    {
      fprintf(file, "\n");
      parameter_first = false;
    }
    if ( parameter_cur->testTag(TAG_HIDDEN) )
      fprintf(file, "//");
    fprintf(file, "  %s %s = %g;\n", VERILOG_KEYWORD_PARAMETER, parameter_cur->getName(), parameter_cur->getValue());
  }
}

//---------------------------------------------------------------------------
void Circuit::writeVerilogModuleInstances(FILE* file, Module* module) const
{
  bool instance_first = true;
  for (int instance_index = 0; instance_index < module->instances->count(); instance_index++)
  {
    Instance* instance_cur = (*module->instances)[instance_index];
    if ( !instance_cur->testTag(TAG_HIDDEN) )
    {
      if (instance_first)
      {
        fprintf(file, "\n");
        instance_first = false;
      }
      int line_length = 2 + 1 + 2;
      char* instance_name = instance_cur->genFullName();
      fprintf(file, "  %s", instance_cur->getModuleName());
      if (instance_name)
        fprintf(file, " %s", instance_name);
      if (instance_cur->getModuleName())
        line_length += strlen(instance_cur->getModuleName());
      if (instance_name)
        line_length += strlen(instance_name);
      delete [] instance_name;
  
      // connection list
      bool connection_first = true;
      for (int connection_index=0; connection_index < instance_cur->connections->count(); connection_index++)
      {
        Connection* connection_cur = (*instance_cur->connections)[connection_index];
        if (connection_first)
        {
          fprintf(file, " (");
          connection_first = false;
        }
        else
          fprintf(file, ", ");
        line_length += 2;
        
        const char* connection_name = connection_cur->getName();
        if (connection_name)
        {
          if (line_length + strlen(connection_name) + 2 > MAX_LINE_LENGTH)
          {
            line_length = 4;
            fprintf(file, "\n    .%s(", connection_name);
          }
          else
            fprintf(file, ".%s(", connection_name);
          line_length += 2 + strlen(connection_name);
        }
        if (connection_cur->nets->count() > 1)
          fprintf(file, "{");
        for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
        {
          Net* net_cur = (*connection_cur->nets)[net_index];
          if ( net_index > 0 )
            fprintf(file, ", ");
            
          char* net_name = NULL;
          if (net_cur)
            net_name = net_cur->genFullName();
          else
            net_name = namecopy("");
  
          if (line_length + strlen(net_name) > MAX_LINE_LENGTH)
          {
            line_length = 4;
            fprintf(file, "\n    %s", net_name);
          }
          else
            fprintf(file, "%s", net_name);
          line_length += 2 + strlen(net_name);
          delete [] net_name;
        }
        if (connection_cur->nets->count() > 1)
          fprintf(file, "}");
        if (connection_name)
          fprintf(file, ")");
      }
      fprintf(file, ");");
      Object* instance_data = (Object*)instance_cur->getData();
      if (instance_data)
        fprintf(file, " // %s, ", instance_data->getName());
      else
        fprintf(file, " // ");
      fprintf(file, "depth=%d, layer=%d", instance_cur->getDepth(), instance_cur->getLayer());
      fprintf(file, "\n");
    }
  }
}

//---------------------------------------------------------------------------
void Circuit::writeVerilogModule(FILE* file, Module* module) const
{
  if ( !module->testTag(TAG_HIDDEN) )
  {
    writeVerilogModuleInterface(file, module);
    writeVerilogModulePorts(file, module);
    writeVerilogModuleParameters(file, module);
    writeVerilogModuleInstances(file, module);
    fprintf(file, "%s\n\n", VERILOG_KEYWORD_ENDMODULE);
  }
}

//---------------------------------------------------------------------------
void Circuit::writeVerilog(const char* file_name, ModuleType module_type) const
{
  if (file_name)
  {
    FILE* file = NULL;
    if (strcmp(file_name, NAME_STDIO) == 0)
      file = stdout;
    else
    {
      if ((file = fopen(file_name, "w")) == NULL)
        Error("opening output file <%s>", file_name);
    }
    for (int module_index = 0; module_index < modules->count(); module_index++)
    {
      Module* module_cur = (*modules)[module_index];
      if (module_cur->getType() == module_type)
        writeVerilogModule(file, module_cur);
    }
    if (file != stdout)
      fclose(file);
  }
}
