#include "messages.h"
#include "sverilog.h"
#include "sv_circuit_imp.h"
#include "properties.h"

const int SIZE_COMPLETION_BLOCK = 4;
const int SIZE_ELEMENT_C2       = 14;
const int SIZE_GATE2            = 4;
const int SIZE_GATE3            = 6;

//---------------------------------------------------------------------------
CircuitImp::CircuitImp(const char* _name, ModuleArray* _modules, ImplementationArray* _implementations,
PropertyArray* _properties, Tag _tag, void* _data) 
: Circuit(_name, _modules, _tag, _data)
{
  // implementations
  if (_implementations)
    implementations = new ImplementationArray(_implementations->count());
  else
    implementations = new ImplementationArray(0);
  addImplementations(_implementations);
  // properties
  property_hash_set = new PropertyHashSet();
  if (_properties)
    properties = new PropertyArray(_properties->count());
  else
    properties = new PropertyArray(0);
  addProperties(_properties);
}

//---------------------------------------------------------------------------
CircuitImp::CircuitImp(const CircuitImp &circuit_imp) 
: Circuit(circuit_imp)
{
  // implementations
  implementations = new ImplementationArray(circuit_imp.implementations->count());
  for (int implementation_index = 0; implementation_index < circuit_imp.implementations->count(); implementation_index++)
  {
    Implementation* implementation_cur = (*circuit_imp.implementations)[implementation_index];
    addImplementation( new Implementation(*implementation_cur) );
  }
  // properties
  property_hash_set = new PropertyHashSet();
  properties = new PropertyArray(circuit_imp.properties->count());
  for (int property_index = 0; property_index < circuit_imp.properties->count(); property_index++)
  {
    Property* property_cur = (*circuit_imp.properties)[property_index];
    addProperty( new Property(*property_cur) );
  }
}

//---------------------------------------------------------------------------
CircuitImp::~CircuitImp()
{
  clean();
  delete implementations; implementations = NULL;
  delete property_hash_set; property_hash_set = NULL;
  delete properties; properties = NULL;
}

//---------------------------------------------------------------------------
void CircuitImp::clean()
{
  implementations->clean(true);
  for (int property_index = 0; property_index < properties->count(); property_index++)
  {
    Property* property_cur = (*properties)[property_index];
    ImplementationArray* table = (ImplementationArray*)property_cur->getData();
    delete table;
  }
  properties->clean(true);
}

//---------------------------------------------------------------------------
void CircuitImp::addImplementation(Implementation* implementation, int position)
{
  if (position < 0)
    implementations->add(implementation);
  else
    implementations->insert(implementation, position);
}

//---------------------------------------------------------------------------
void CircuitImp::addImplementations(ImplementationArray* _implementations)
{
  if (_implementations)
  for (int implementation_index = 0; implementation_index < _implementations->count(); implementation_index++)
  {
    Implementation* implementation_cur = (*_implementations)[implementation_index];
    addImplementation(implementation_cur);
  }
}

//---------------------------------------------------------------------------
void CircuitImp::removeImplementation(Implementation* implementation)
{
  implementations->remove(implementation, true);
}

//---------------------------------------------------------------------------
Implementation* CircuitImp::findImplementation(const char* implementation_name) const
{
  Implementation* implementation_mask = new Implementation(implementation_name);
  int implementation_index = implementations->find(implementation_mask, 0);
  delete implementation_mask;
  if (implementation_index >= 0)
    return  (*implementations)[implementation_index];
  else
    return NULL;
}

//---------------------------------------------------------------------------
void CircuitImp::addProperty(Property* property, int position)
{
  if (position < 0)
    properties->add(property);
  else
    properties->insert(property, position);
  property_hash_set->add(property);
}

//---------------------------------------------------------------------------
void CircuitImp::addProperties(PropertyArray* _properties)
{
  if (_properties)
  for (int property_index = 0; property_index < _properties->count(); property_index++)
  {
    Property* property_cur = (*_properties)[property_index];
    addProperty(property_cur);
  }
}

//---------------------------------------------------------------------------
void CircuitImp::removeProperty(Property* property)
{
  properties->remove(property, true);
  property_hash_set->remove(property);
}

//---------------------------------------------------------------------------
Property* CircuitImp::findProperty(const char* property_name) const
{
  Property* property_mask = new Property(property_name);
  Property* property_cur = property_hash_set->find(property_mask);
  delete property_mask;
  return property_cur;
}

//---------------------------------------------------------------------------
void CircuitImp::sortImplementations()
{
  for (int property_index = 0; property_index < properties->count(); property_index++)
  {
    Property* property_cur = (*properties)[property_index];
    for (int implementation_prot_index = 0; implementation_prot_index < implementations->count(); implementation_prot_index++)
    {
      Implementation* implementation_prot = (*implementations)[implementation_prot_index];
      bool implementation_found = false;
      for (int implementation_index = implementation_prot_index; implementation_index < property_cur->implementations->count(); implementation_index++)
      {
        Implementation* implementation_cur = (*property_cur->implementations)[implementation_index];
        if ( strcmp(implementation_prot->getName(), implementation_cur->getName()) == 0 )
        {
          if (implementation_prot_index != implementation_index)
          {
            (*property_cur->implementations)[implementation_index] = (*property_cur->implementations)[implementation_prot_index];
            (*property_cur->implementations)[implementation_prot_index] = implementation_cur;
          }
            implementation_found = true;
            break;
        }
      }
      if ( !implementation_found )
      {
        Implementation* implementation_new = new Implementation(implementation_prot->getName(), implementation_prot->getType());
        property_cur->addImplementation(implementation_new, implementation_prot_index);
      }
    }
  }
}

//---------------------------------------------------------------------------
void CircuitImp::calcModuleImplementations(Module* module)
{
  if ( !module->testTag(TAG_PROCESSED) && (module->getType() == MODULE_TYPE_COMPOUND) )
  {
    module->addTag(TAG_PROCESSED);
    Property* property_new = findProperty(module->getName());
    if (property_new == NULL)
    {
      property_new = new Property(module->getName());
      for (int implementation_index = 0; implementation_index < implementations->count(); implementation_index++)
      {
        Implementation* implementation_cur = (*implementations)[implementation_index];
        Implementation* implementation_new = new Implementation(implementation_cur->getName(), implementation_cur->getType(), 0);
        property_new->addImplementation(implementation_new, implementation_index);
      }
      addProperty(property_new);
      for (int instance_index = 0; instance_index < module->instances->count(); instance_index++)
      {
        Instance* instance_cur = (*module->instances)[instance_index];
        Module* module_prot = findModule(instance_cur->getModuleName());
        if (module_prot == NULL)
        {
          char* instance_name = instance_cur->genFullName();
          Error("no module <%s> found for instance <%s> in module <%s>", instance_cur->getModuleName(), instance_name, name);
          delete [] instance_name;
        }
        else
        {
          calcModuleImplementations(module_prot);
          Property* property_cur = findProperty(module_prot->getName());
          if (property_cur == NULL)
          {
            char* instance_name = instance_cur->genFullName();
            Error("cannot find implementations for module <%s> which is necessary for instance <%s> in module <%s>", 
              module_prot->getName(), instance_name, module->getName());
            delete [] instance_name;
          }
          else
          {
            for (int implementation_index = 0; implementation_index < property_new->implementations->count(); implementation_index++)
            {
              Implementation* implementation_new = (*property_new->implementations)[implementation_index];
              Implementation* implementation_cur = (*property_cur->implementations)[implementation_index];
              int value_new = implementation_new->getValue();
              int value_cur = implementation_cur->getValue();
              if ( (value_new < 0) || (value_cur < 0) )
                implementation_new->setValue(-1);
              else
                implementation_new->setValue(value_new + value_cur);
            }
          }
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void CircuitImp::calcImplementations()
{
  if ( !top ) calcTop();
  if ( !top )
    Warning("cannot calc implementations because the top module is not defined");
  else
  {
    for (int module_index = 0; module_index < modules->count(); module_index++)
      (*modules)[module_index]->removeTag(TAG_PROCESSED);
    calcModuleImplementations(top);
    Property* property_top = findProperty(top->getName());
    for (int implementation_index = 0; implementation_index < implementations->count(); implementation_index++)
    {
      Implementation* implementation_cur = (*implementations)[implementation_index];
      Implementation* implementation_top = (*property_top->implementations)[implementation_index];
      implementation_cur->setValue(implementation_top->getValue());
    }
  }
}

//---------------------------------------------------------------------------
int CircuitImp::calcModuleAllForks(Module* module, const char* forks_name)
{
  int result = 0;
  if (module->getType() == MODULE_TYPE_COMPOUND)
  {
    Property* property_cur = findProperty(module->getName());
    if (module->testTag(TAG_PROCESSED))
    {
      Implementation* implementation_forks = property_cur->findImplementation(forks_name);
      result = implementation_forks->getValue();
    }
    else
    {
      for (int instance_index = 0; instance_index < module->instances->count(); instance_index++)
      {
        Instance* instance_cur = (*module->instances)[instance_index];
        Module* module_prot = findModule(instance_cur->getModuleName());
        if (module_prot->getType() == MODULE_TYPE_COMPOUND)
          result += calcModuleAllForks(module_prot, forks_name);
        else
        {
          for (int connection_index =  0; connection_index < instance_cur->connections->count(); connection_index++)
          {
            Connection* connection_cur = (*instance_cur->connections)[connection_index];
            if ( (connection_cur->getType() == CONNECTION_TYPE_INPUT)
              || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
            {
              for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
                if ( (*connection_cur->nets)[net_index] != NULL )
                  result++;
            }
          }
        }
      }      
      Implementation* implementation_forks = new Implementation(forks_name, IMPLEMENTATION_TYPE_FORKS, result);
      property_cur->addImplementation(implementation_forks);
    }
  }
  return result;
}

//---------------------------------------------------------------------------
void CircuitImp::calcAllForks(const char* forks_name)
{
  if ( !top ) calcTop();
  if ( !top )
    Warning("cannot calc isochronic forks because the top module is not defined");
  else
  {
    for (int module_index = 0; module_index < modules->count(); module_index++)
      (*modules)[module_index]->removeTag(TAG_PROCESSED);
    int forks_num = calcModuleAllForks(top, forks_name);
    for (int port_index = 0; port_index < top->ports->count(); port_index++)
    {
      Port* port_cur = (*top->ports)[port_index];
      if ( (port_cur->getType() == PORT_TYPE_OUTPUT)
        || (port_cur->getType() == PORT_TYPE_INOUT) )
        forks_num += port_cur->nets->count();
    }
    Implementation* implementation_forks = new Implementation(forks_name, IMPLEMENTATION_TYPE_FORKS, forks_num);
    addImplementation(implementation_forks);
  }
}
                                                                                                                                                    
//---------------------------------------------------------------------------
void CircuitImp::findElementC_x2x3(int input_num, int &x2, int &x3)
{                                                                                                                                                   
  if (input_num % 3 == 1)
    x3 = (input_num-2) / 3;
  else
    x3 = input_num / 3;
  x2 = (input_num - x3*3) / 2;
}

//---------------------------------------------------------------------------
int CircuitImp::calcElementC(int input_num)
{
  int result = 0;
  if (input_num > 2)
  {
    int layer_input_num = input_num;
    while (layer_input_num > 1)
    {
      int layer_gate2_num = 0;
      int layer_gate3_num = 0;
      findElementC_x2x3(layer_input_num, layer_gate2_num, layer_gate3_num);
      result += SIZE_GATE2 * layer_gate2_num;
      result += SIZE_GATE3 * layer_gate3_num;
      layer_input_num = layer_gate2_num + layer_gate3_num;
    }
    result *= 2;
  }
  result += SIZE_ELEMENT_C2;
  return result;
}

//---------------------------------------------------------------------------
int CircuitImp::calcModuleCompletion(Module* module, const char* completion_name, bool full)
{
  int result = 0;
  int input_num = 1;
  if (module->getType() == MODULE_TYPE_COMPOUND)
  {
    Property* property_cur = findProperty(module->getName());
    if (module->testTag(TAG_PROCESSED))
    {
      Implementation* implementation_completion = property_cur->findImplementation(completion_name);
      result = implementation_completion->getValue();
    }
    else
    {
      for (int instance_index = 0; instance_index < module->instances->count(); instance_index++)
      {
        Instance* instance_cur = (*module->instances)[instance_index];
        Module* module_prot = findModule(instance_cur->getModuleName());
        if (module_prot->getType() == MODULE_TYPE_COMPOUND)
        {
          result += calcModuleCompletion(module_prot, completion_name, full);
          input_num++;
        }
        else
        {
          if (full)
          for (int connection_index =  0; connection_index < instance_cur->connections->count(); connection_index++)
          {
            Connection* connection_cur = (*instance_cur->connections)[connection_index];
            if ( (connection_cur->getType() == CONNECTION_TYPE_OUTPUT)
              || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
            {
              for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
                if ( (*connection_cur->nets)[net_index] != NULL )
                {
                  result += SIZE_COMPLETION_BLOCK;
                  input_num++;
                }
            }
          }
        }
      }
      result += calcElementC(input_num);
      Implementation* implementation_completion = new Implementation(completion_name, IMPLEMENTATION_TYPE_COMPLETION, result);
      property_cur->addImplementation(implementation_completion);
    }
  }
  return result;
}

//---------------------------------------------------------------------------
void CircuitImp::calcCompletion(const char* completion_name, bool full)
{
  if ( !top ) calcTop();
  if ( !top )
    Warning("cannot calc full completion detection logic because the top module is not defined");
  else
  {
    for (int module_index = 0; module_index < modules->count(); module_index++)
      (*modules)[module_index]->removeTag(TAG_PROCESSED);
    int completion_num = calcModuleCompletion(top, completion_name, full);
    Implementation* implementation_completion = new Implementation(completion_name, IMPLEMENTATION_TYPE_COMPLETION, completion_num);
    addImplementation(implementation_completion);
  }
}

//---------------------------------------------------------------------------
void CircuitImp::readProperties(const char* file_names)
{
  if (file_names)
  {
    int name_index = 0;
    while (char* file_name = names2name(file_names, name_index++))
    {
      if (strcmp(file_name, NAME_STDIO) == 0)
        properties_in = stdin;
      else
        if ((properties_in = fopen(file_name, "r")) == NULL)
          Error("opening input file <%s>", file_name);
      properties_parse(this, file_name);
      if (properties_in != stdin)
        fclose(properties_in);
      delete [] file_name;
    }
  }
}
  
//---------------------------------------------------------------------------
void CircuitImp::writeProperties(const char* file_name) const
{
  if (file_name)
  {
    FILE* file = NULL;
    if (strcmp(file_name, NAME_STDIO) == 0)
      file = stdout;
    else
    {
      if ((file = fopen(file_name, "w")) == NULL)
        Error("opening output file <%s>", file_name);
    }
    for (int property_index = 0; property_index < properties->count(); property_index++)
    {
      Property* property_cur = (*properties)[property_index];
      printf("%s\n", property_cur->getName());
      for (int implementation_index = 0; implementation_index < property_cur->implementations->count(); implementation_index++)
      {
        Implementation* implementation_cur = (*property_cur->implementations)[implementation_index];
        printf("  %s = %d (%d)\n", implementation_cur->getName(), implementation_cur->getValue(), implementation_cur->getType());
      }
      // Optimisation table
      ImplementationArray* table = (ImplementationArray*) property_cur->getData();
      if (table)
      {
        printf("Table for %s\n", property_cur->getName());
        for (int implementation_index = 0; implementation_index < table->count(); implementation_index++)
        {
          Implementation* implementation_cur = (*table)[implementation_index];
          if (implementation_cur)
            printf("  %x: %s = %d (%d)\n", implementation_index, implementation_cur->getName(), implementation_cur->getValue(), implementation_cur->getType());
        }
      }
    }
    if (file != stdout)
      fclose(file);
  }
}
  
//---------------------------------------------------------------------------
void CircuitImp::writeImplementations(FILE* file) const
{
  bool implementation_first = true;
  for (int implementation_index = 0; implementation_index < implementations->count(); implementation_index++)
  {
    Implementation* implementation_cur = (*implementations)[implementation_index];
    if (implementation_first)
    {
      fprintf(file, "//===========================================================================\n");
      fprintf(file, "// Implementaions of the circuit with top module <%s>:\n", top->getName());
      implementation_first = false;
    }
    fprintf(file, "//   %s = %d", implementation_cur->getName(), implementation_cur->getValue());
    switch (implementation_cur->getType())
    {
      case IMPLEMENTATION_TYPE_NONE:
        fprintf(file, "  - none-indicating implementation");
        break;
      case IMPLEMENTATION_TYPE_FULL:
        fprintf(file, "  - fully indicating implementation");
        break;
      case IMPLEMENTATION_TYPE_OPT:
        fprintf(file, "  - optimised fully indicating implementation");
        break;
      case IMPLEMENTATION_TYPE_FORKS:
        fprintf(file, "  - isochronic forks");
        break;
      case IMPLEMENTATION_TYPE_COMPLETION:
        fprintf(file, "  - completion detection logic");
        break;
      default:
        fprintf(file, "  - undefined implementation");
        break;
    }
    fprintf(file, "\n");
  }
  if ( !implementation_first)
    fprintf(file, "\n");
}

//---------------------------------------------------------------------------
void CircuitImp::writeModuleImplementations(FILE* file, Module* module) const
{
  Property* property_cur = findProperty(module->getName());
  if (property_cur)
  {
    bool implementation_first = true;
    for (int implementation_index = 0; implementation_index < property_cur->implementations->count(); implementation_index++)
    {
      Implementation* implementation_cur = (*property_cur->implementations)[implementation_index];
      if (implementation_first)
      {
        fprintf(file, "//---------------------------------------------------------------------------\n");
        fprintf(file, "// Implementaions of module <%s>:\n", module->getName());
        implementation_first = false;
      }
      fprintf(file, "//   %s = %d", implementation_cur->getName(), implementation_cur->getValue());
      switch (implementation_cur->getType())
      {
        case IMPLEMENTATION_TYPE_NONE:
          fprintf(file, "  - none-indicating implementation");
          break;
        case IMPLEMENTATION_TYPE_FULL:
          fprintf(file, "  - fully indicating implementation");
          break;
        case IMPLEMENTATION_TYPE_OPT:
          fprintf(file, "  - otimised fully indicating implementation");
          break;
        case IMPLEMENTATION_TYPE_FORKS:
          fprintf(file, "  - isochronic forks");
          break;
        case IMPLEMENTATION_TYPE_COMPLETION:
          fprintf(file, "  - completion detection logic");
          break;
        default:
          fprintf(file, "  - undefined implementation");
          break;
      }
      fprintf(file, "\n");
    }
  }
}

//---------------------------------------------------------------------------
void CircuitImp::writeVerilogImp(const char* file_name, ModuleType module_type, int statistics_level) const
{
  if (file_name)
  {
    FILE* file = NULL;
    if (strcmp(file_name, NAME_STDIO) == 0)
      file = stdout;
    else
    {
      if ((file = fopen(file_name, "w")) == NULL)
        Error("opening output file <%s>", file_name);
    }
    if (statistics_level > 0)
      writeImplementations(file); 
    for (int module_index = 0; module_index < modules->count(); module_index++)
    {
      Module* module_cur = (*modules)[module_index];
      if (module_cur->getType() == module_type)
      {
        if (statistics_level > 1)
          writeModuleImplementations(file, module_cur); 
        writeVerilogModule(file, module_cur);
      }
    }
    if (file != stdout)
      fclose(file);
  }
}
