#include "messages.h"
#include "sverilog.h"
#include "sv_object.h"
#include "sv_region.h"
#include "sv_indicator.h"
#include "sv_circuit_imp_opt.h"
#include "properties.h"

//---------------------------------------------------------------------------
const Tag TAG_SET = TAG_BIT_USER_1;
const Tag TAG_ORPHAN = TAG_BIT_USER_2;

//---------------------------------------------------------------------------
CircuitImpOpt::CircuitImpOpt(const char* _name, ModuleArray* _modules, ImplementationArray* _implementations,
PropertyArray* _properties, unsigned _iteration_max, Tag _tag, void* _data)
: CircuitImp(_name, _modules, _implementations, _properties, _tag, _data), optimised_implementation_name(NULL), 
  iteration_max(_iteration_max), indent(0), region_optimisation(true)
{
}

//---------------------------------------------------------------------------
CircuitImpOpt::~CircuitImpOpt()
{
  delete [] optimised_implementation_name; optimised_implementation_name = NULL;
}

//---------------------------------------------------------------------------
int CircuitImpOpt::calcVariationCode(const char* variation_name, Module* module) const
{
  int variation_code = 0;
  char* buf = namecopy(variation_name);
  char* port_name_start = strchr(buf, ')');
  if (port_name_start)
    port_name_start[0]=',';
  port_name_start = strchr(buf, '(');
  while (port_name_start != NULL)
  {
    port_name_start++;
    char* port_name_end = strchr(port_name_start, ',');
    if (port_name_end != NULL)
    {
      port_name_end[0]='\0';
      int input_index = -1;
      for (int port_index = 0; port_index < module->ports->count(); port_index++)
      {
        Port* port_cur = (*module->ports)[port_index];
        if ( (port_cur->getType() == PORT_TYPE_INPUT)
          || (port_cur->getType() == PORT_TYPE_INOUT) )
          input_index++;
        if (strcmp(port_cur->getName(), port_name_start) == 0)
        {
          variation_code |= (1 << input_index);
          break;
        }
      }
      port_name_start = port_name_end;
    }
    else
      port_name_start = NULL;
  }
  delete [] buf;
  return variation_code;
}

//---------------------------------------------------------------------------
void CircuitImpOpt::calcTables(const char* implementation_names)
{
  char* implementation_names_tmp = NULL;
  if (implementation_names != NULL)
    implementation_names_tmp = namecopy(implementation_names);
  else
  {
    for (int implementation_index = 0; implementation_index < implementations->count(); implementation_index++)
    {
      Implementation* implementation_cur = (*implementations)[implementation_index];
      if ( (implementation_cur->getType() == IMPLEMENTATION_TYPE_PARTIAL)
        || (implementation_cur->getType() == IMPLEMENTATION_TYPE_FULL)
        || (implementation_cur->getType() == IMPLEMENTATION_TYPE_NONE) )
      {
        if (implementation_names_tmp == NULL)
          implementation_names_tmp = namecopy(implementation_cur->getName());
        else
        {
          char* s = implementation_names_tmp;
          implementation_names_tmp = name2pattern("%s,%s", s, implementation_cur->getName());
          delete [] s;
        }
      }
    }
  }
  setOptimisedImplementationName(implementation_names_tmp);

  char* implementation_names_opt = name2pattern(",%s,", implementation_names_tmp);
  delete [] implementation_names_tmp;
  
  for (int property_index = 0; property_index < properties->count(); property_index++)
  {
    Property* property_cur = (*properties)[property_index];
    ImplementationArray* table = NULL;
    Module* module_cur = findModule(property_cur->getName());
    if (module_cur == NULL)
      Warning("property <%s> cannot be associated with a module", property_cur->getName());
    else
    {
      int input_count = module_cur->getPortCount(PORT_TYPE_INPUT) + module_cur->getPortCount(PORT_TYPE_INOUT);
      int code_none = 0;
      int code_all = (1 << input_count) - 1;
      int implementation_count = (1 << input_count);
      table = new ImplementationArray(implementation_count);
      for (int implementation_index = 0; implementation_index < implementation_count; implementation_index++)
        table->add(NULL);
      for (int implementation_index = 0; implementation_index < property_cur->implementations->count(); implementation_index++)
      {
        Implementation* implementation_cur = (*property_cur->implementations)[implementation_index];
        int code_cur = -1;
        if (implementation_cur->getValue() >= 0)
        {
          char* implementation_name_opt = namecopy(implementation_cur->getName());
          char* s = strchr(implementation_name_opt, '(');
          if (s) s[0] = '\0';
          if ( (implementation_names_opt == NULL) 
            || (strstr(implementation_names_opt, implementation_name_opt) != NULL) )
          {
            // none-indicating
            if (implementation_cur->getType() == IMPLEMENTATION_TYPE_NONE)
              code_cur = code_none;
            else
            // partially-indicating
            if (implementation_cur->getType() == IMPLEMENTATION_TYPE_PARTIAL)
              code_cur = calcVariationCode(implementation_cur->getName(), module_cur);
            else
            // fully-indicating
            if (implementation_cur->getType() == IMPLEMENTATION_TYPE_FULL)
              code_cur = code_all;
  
            if(code_cur >= 0)
            {
              Implementation*& table_cur = (*table)[code_cur];
              if ( (table_cur == NULL) || (implementation_cur->getValue() < table_cur->getValue()) )
                table_cur = implementation_cur;
            }
          }
          delete [] implementation_name_opt;
        }
      }
    }
    ImplementationArray* table_old = (ImplementationArray*)property_cur->getData();
    delete table_old; table_old = NULL;
    property_cur->setData(table);
  }
  delete [] implementation_names_opt;
}

//---------------------------------------------------------------------------
void CircuitImpOpt::optimiseTables()
{
  for (int property_index = 0; property_index < properties->count(); property_index++)
  {
    Property* property_cur = (*properties)[property_index];
    ImplementationArray* table = (ImplementationArray*)property_cur->getData();
    for (int implementation_index_1 = 0; implementation_index_1 < table->count(); implementation_index_1++)
    {
      Implementation*& implementation_1 = (*table)[implementation_index_1];
      for (int implementation_index_2 = implementation_index_1+1; implementation_index_2 < table->count(); implementation_index_2++)
      {
        if( (implementation_index_2 & implementation_index_1) == implementation_index_1 )
        {
          Implementation*& implementation_2 = (*table)[implementation_index_2];
          if (implementation_2)
            if ( !implementation_1 || (implementation_2->getValue() < implementation_1->getValue()) )
              implementation_1 = implementation_2;
        }
      }      
    }
    // check the optimisation table
    for (int implementation_index = 0; implementation_index < table->count(); implementation_index++)
    {
      Implementation*& implementation_cur = (*table)[implementation_index];
      if (implementation_cur == NULL)
        Error("cannot build optimisation table for module <%s>", property_cur->getName());
    }
  }
}

//---------------------------------------------------------------------------
int CircuitImpOpt::calcSolutionValue(Region* region)
{
  int result = 0;
  for (int instance_index = 0; instance_index < region->instances->count(); instance_index++)
  {
    Instance* instance_cur = (*region->instances)[instance_index];
    Property* property_cur = findProperty(instance_cur->getModuleName());
    ImplementationArray* table_cur = (ImplementationArray*)property_cur->getData();
    if (table_cur)
    {
      unsigned* solution = region->getSolution();
      int table_index = solution[instance_index];
      Implementation* implementation_cur = (*table_cur)[table_index];
      result += implementation_cur->getValue();
    }
  }
  return result;
}

//---------------------------------------------------------------------------
bool CircuitImpOpt::findRegionImplementation(Region* region, int net_index)
{
  if (net_index < region->nets->count())
  {
    Net* net_cur = (*region->nets)[net_index];
    IndicatorArray* indicators = (IndicatorArray*)net_cur->getData();
    for (int  indicator_index = 0; indicator_index < indicators->count(); indicator_index++)
    {
      Indicator* indicator_cur = (*indicators)[indicator_index];
      int instance_index = indicator_cur->getObjectIndex();
      int input_index = indicator_cur->getInputIndex();
      region->addSolution(instance_index, input_index);
      if (findRegionImplementation(region, net_index + 1))
        break;
      region->removeSolution(instance_index, input_index);
    }
  }
  else
  {
    region->incIteration();
    int value = calcSolutionValue(region);
    if ( (region->getBestValue() < 0) || (region->getBestValue() > value) )
    {
      region->saveSolution();
      region->setBestValue(value);
    }
  }
  return (region->getIteration() >= iteration_max);
}

//---------------------------------------------------------------------------
int CircuitImpOpt::calcRegionImplementation(Region* region)
{
  int result = 0;
  if ( (region->nets->count() > 0) || (region->instances->count() > 0) )
  {
    // preparation
    double variants_count = 1;
    indent += 2;
    for (int net_index = 0; net_index < region->nets->count(); net_index++)
    {
      Net* net_cur = (*region->nets)[net_index];
      IndicatorArray* indicators = (IndicatorArray*)net_cur->getData();
      variants_count *= indicators->count();
    }
    Text(indent, "region <%s>: %d unacknowledged net(s) with %.8lg variant(s) of indication by %d instance(s)\n", 
      region->getName(), region->nets->count(), variants_count, region->instances->count());
    // calculation  
    region->initSolution();
    findRegionImplementation(region, 0);
    region->loadSolution();
    result += calcSolutionValue(region);
    if (region->getIteration() < variants_count)
      Text(indent, "breaking after concidering  %d variant(s)\n", region->getIteration());
    indent -= 2;
  }
  return result;
}

//---------------------------------------------------------------------------
void CircuitImpOpt::calcRegionData(Region* region)
{
  for (int region_instance_index = 0; region_instance_index < region->instances->count(); region_instance_index++)
  {
    Instance* instance_cur = (*region->instances)[region_instance_index];
    int input_index = -1;
    for (int connection_index = 0; connection_index < instance_cur->connections->count(); connection_index++)
    {
      Connection* connection_cur = (*instance_cur->connections)[connection_index]; 
      if ( (connection_cur->getType() == CONNECTION_TYPE_INPUT) 
        || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
      {
        input_index++;
        for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
        {
          Net* net_cur = (*connection_cur->nets)[net_index];
          if (net_cur && net_cur->testTag(TAG_SET))
          {
            // instance preset
            IndicatorArray* instance_indicators = (IndicatorArray*)instance_cur->getData();
            if (instance_indicators == NULL)
            {
              instance_indicators = new IndicatorArray(1);
              instance_cur->setData(instance_indicators);
            }
            char* instance_indicator_name = name2pattern("%s:%s", net_cur->getName(), connection_cur->getName());
            int region_net_index = region->nets->locate(net_cur);
            Indicator* instance_indicator_new = new Indicator(instance_indicator_name, region_net_index, input_index);
            delete [] instance_indicator_name;
            instance_indicators->add(instance_indicator_new);
            // net postset
            IndicatorArray* net_indicators = (IndicatorArray*)net_cur->getData();
            if (net_indicators == NULL)
            {
              net_indicators = new IndicatorArray(1);
              net_cur->setData(net_indicators);
            }
            char* net_indicator_name = name2pattern("%s:%s", instance_cur->getName(), connection_cur->getName());
            Indicator* net_indicator_new = new Indicator(net_indicator_name, region_instance_index, input_index);
            delete [] net_indicator_name;
            net_indicators->add(net_indicator_new);
          }
        } 
      }
    }
  }
}

//---------------------------------------------------------------------------
void CircuitImpOpt::saveRegionData(Region* region)
{
  for (int instance_index = 0; instance_index < region->instances->count(); instance_index++)
  {
    Instance* instance_cur = (*region->instances)[instance_index];
    Property* property_cur = findProperty(instance_cur->getModuleName());
    ImplementationArray* table_cur = (ImplementationArray*)property_cur->getData();
    unsigned* solution = region->getSolution();
    if (table_cur)
    {
      int table_index = solution[instance_index];
      Implementation* implementation_cur = (*table_cur)[table_index];
      instance_cur->setData(implementation_cur);
    }
  }
}

//---------------------------------------------------------------------------
void CircuitImpOpt::cleanRegionData(Region* region)
{
  for (int instance_index = 0; instance_index < region->instances->count(); instance_index++)
  {
    Instance* instance_cur = (*region->instances)[instance_index];
    IndicatorArray* instance_indicators = (IndicatorArray*)instance_cur->getData();
    if (instance_indicators) instance_indicators->clean(true);
    delete instance_indicators; instance_indicators = NULL;
  }
  for (int net_index = 0; net_index < region->nets->count(); net_index++)
  {
    Net* net_cur = (*region->nets)[net_index];
    IndicatorArray* net_indicators = (IndicatorArray*)net_cur->getData();
    if (net_indicators) net_indicators->clean(true);
    delete net_indicators; net_indicators = NULL;
  }
}

//---------------------------------------------------------------------------
void CircuitImpOpt::findRegionInstance(Set* set, Region* region, Instance* instance)
{
  if ( !instance->testTag(TAG_PROCESSED) )
  {
    instance->addTag(TAG_PROCESSED);
    region->addInstance(instance);
    for (int connection_index = 0; connection_index < instance->connections->count(); connection_index++)
    {
      Connection* connection_cur = (*instance->connections)[connection_index];
      if ( (connection_cur->getType() == CONNECTION_TYPE_INPUT)
        || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
      {
        for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
        {
          Net* net_cur = (*connection_cur->nets)[net_index];
          if (net_cur && net_cur->testTag(TAG_SET) )
            findRegionNet(set, region, net_cur);
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void CircuitImpOpt::findRegionNet(Set* set, Region* region, Net* net)
{
  if ( !net->testTag(TAG_PROCESSED) )
  {
    net->addTag(TAG_PROCESSED);
    region->addNet(net);
    for (int instance_index = 0; instance_index < net->postset->count(); instance_index++)
    {
      Instance* instance_cur = (*net->postset)[instance_index];
      if (instance_cur->testTag(TAG_SET))
        findRegionInstance(set, region, instance_cur);
    }
  }
}

//---------------------------------------------------------------------------
int CircuitImpOpt::calcRegionwiseImplementation(Set* set)
{
  int result = 0;
//  indent += 2;
  // prepare set data
  for (int instance_index = 0; instance_index < set->instances->count(); instance_index++)
  {
    Instance* instance_cur = (*set->instances)[instance_index];
    instance_cur->removeTag(TAG_PROCESSED);
    instance_cur->addTag(TAG_SET);
  }
  for (int net_index = 0; net_index < set->nets->count(); net_index++)
  {
    Net* net_cur = (*set->nets)[net_index];
    net_cur->removeTag(TAG_PROCESSED);
    net_cur->addTag(TAG_SET);
  }
  if (region_optimisation)
  {
    // calc set regions
    int region_index = -1;
    for (int net_index = 0; net_index < set->nets->count(); net_index++)
    {
      Net* net_cur = (*set->nets)[net_index];
      if ( !net_cur->testTag(TAG_PROCESSED) )
      {
        region_index++;
        char* region_name = name2pattern("%s-%.2d", set->getName(), region_index);
        Region* region = new Region(region_name);
        delete [] region_name;
        findRegionNet(set, region, net_cur);
        calcRegionData(region);
        result += calcRegionImplementation(region);
        cleanRegionData(region);
        saveRegionData(region);
        delete region;
      }
    }
    // final region
    region_index++;
    char* region_name = name2pattern("%s-%.2d", set->getName(), region_index);
    Region* region = new Region(region_name);
    delete [] region_name;
    for (int instance_index = 0; instance_index < set->instances->count(); instance_index++)
    {
      Instance* instance_cur = (*set->instances)[instance_index];
      if ( !instance_cur->testTag(TAG_PROCESSED) )
        region->addInstance(instance_cur);
    }
    calcRegionData(region);
    result += calcRegionImplementation(region);
    cleanRegionData(region);
    saveRegionData(region);
    delete region;
  }
  else
  {
    Region* region = new Region(set->getName(), set->instances, set->nets);
    calcRegionData(region);
    result += calcRegionImplementation(region);
    cleanRegionData(region);
    saveRegionData(region);
    delete region;
  }
  // clean up
  for (int instance_index = 0; instance_index < set->instances->count(); instance_index++)
  {
    Instance* instance_cur = (*set->instances)[instance_index];
    instance_cur->removeTag(TAG_PROCESSED);
    instance_cur->removeTag(TAG_SET);
  }
  for (int net_index = 0; net_index < set->nets->count(); net_index++)
  {
    Net* net_cur = (*set->nets)[net_index];
    net_cur->removeTag(TAG_PROCESSED);
    net_cur->removeTag(TAG_SET);
  }
//  indent -= 2;
  return result;
}

//---------------------------------------------------------------------------
int CircuitImpOpt::calcModuleOptImplementation(Module* module)
{
  int result = 0;
  if(module->getType() == MODULE_TYPE_COMPOUND)
  {
    indent += 2;
    if (module->testTag(TAG_PROCESSED))
    {
      Property* property_cur = findProperty(module->getName());
      if (property_cur)
      {
        Implementation* implementation_opt = property_cur->findImplementation(optimised_implementation_name);
        if (implementation_opt)
          result = implementation_opt->getValue();
      }
    }
    else
    {
      Text(indent, "processing module <%s>\n", module->getName());
      module->addTag(TAG_PROCESSED);
      // build set of instances
      Set* set = new Set(module->getName());
      for (int instance_index = 0; instance_index < module->instances->count(); instance_index++)
      {
        Instance* instance_cur = (*module->instances)[instance_index];
        Module* module_prot = findModule(instance_cur->getModuleName());
        if (module_prot->getType() == MODULE_TYPE_COMPOUND)
          result += calcModuleOptImplementation(module_prot);
        else
          set->addInstance(instance_cur);
      }
      // build set of nets
      for (int port_index = 0; port_index < module->ports->count(); port_index++)
      {
        Port* port_cur = (*module->ports)[port_index];
        if ( !port_cur->testTag(TAG_HIDDEN) )        
        if ( (port_cur->getType() == PORT_TYPE_INPUT) || (port_cur->getType() == PORT_TYPE_WIRE) )
        {
          for (int net_index = 0; net_index < port_cur->nets->count(); net_index++)
          {
            Net* net_cur = (*port_cur->nets)[net_index];
            bool net_is_indicated = false;
            for(int postset_index = 0; postset_index < net_cur->postset->count(); postset_index++)
            {
              Instance* instance_cur = (*net_cur->postset)[postset_index];
              Module* module_prot = findModule(instance_cur->getModuleName());
              if (module_prot->getType() == MODULE_TYPE_COMPOUND)
              {
                net_is_indicated = true;
                break;
              }                
            }
// Message("<%s> %d", net_cur->genFullName(), net_cur->postset->count());
            if ( !net_is_indicated && (net_cur->postset->count() > 0) )
              set->addNet(net_cur);
          }
        }
      }
      result += calcRegionwiseImplementation(set);
      Property* property_cur = findProperty(module->getName());
      Implementation* implementation_opt = new Implementation(optimised_implementation_name, IMPLEMENTATION_TYPE_OPT, result);
      property_cur->addImplementation(implementation_opt);
      delete set;
    }
    indent -= 2;
  }
  return result;
}

//---------------------------------------------------------------------------
void CircuitImpOpt::calcOptImplementation()
{
  if ( !top ) calcTop();
  if ( !top )
    Warning("cannot calc implementations because the top module is not defined");
  else
  {
    for (int module_index = 0; module_index < modules->count(); module_index++)
      (*modules)[module_index]->removeTag(TAG_PROCESSED);
    indent = 0;
    int value_opt = calcModuleOptImplementation(top);
    Implementation* implementation_opt = new Implementation(optimised_implementation_name, IMPLEMENTATION_TYPE_OPT, value_opt);
    addImplementation(implementation_opt);
  }
}

//---------------------------------------------------------------------------
int CircuitImpOpt::calcModuleOptForks(Module* module, const char* forks_name)
{
  int result = 0;
  if (module->getType() == MODULE_TYPE_COMPOUND)
  {
    Property* property_cur = findProperty(module->getName());
    if (module->testTag(TAG_PROCESSED))
    {
      Implementation* implementation_forks = property_cur->findImplementation(forks_name);
      result = implementation_forks->getValue();
    }
    else
    {
      for (int instance_index = 0; instance_index < module->instances->count(); instance_index++)
      {
        Instance* instance_cur = (*module->instances)[instance_index];
        Module* module_prot = findModule(instance_cur->getModuleName());
        if (module_prot->getType() == MODULE_TYPE_COMPOUND)
          result += calcModuleOptForks(module_prot, forks_name);
        else
        {
          int input_count = module_prot->getPortCount(PORT_TYPE_INPUT) + module_prot->getPortCount(PORT_TYPE_INOUT);
          int code_all = (1 << input_count) - 1;
          int code_cur = 0;
          Implementation* implementation_cur = (Implementation*)instance_cur->getData();
          if (implementation_cur->getType() == IMPLEMENTATION_TYPE_PARTIAL)
            code_cur = calcVariationCode(implementation_cur->getName(), module_prot);
          else
          if (implementation_cur->getType() == IMPLEMENTATION_TYPE_FULL)
            code_cur = code_all;
          int input_index = -1;
          for (int connection_index =  0; connection_index < instance_cur->connections->count(); connection_index++)
          {
            Connection* connection_cur = (*instance_cur->connections)[connection_index];
            if ( (connection_cur->getType() == CONNECTION_TYPE_INPUT)
              || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
            {
              input_index++;
              int input_code = (1 << input_index);
              if ((code_cur & input_code) != input_code)
              {
                for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
                  if ( (*connection_cur->nets)[net_index] != NULL )
                    result++;
              }
            }
          }
        }
      }      
      Implementation* implementation_forks = new Implementation(forks_name, IMPLEMENTATION_TYPE_FORKS, result);
      property_cur->addImplementation(implementation_forks);
    }
  }
  return result;
}

//---------------------------------------------------------------------------
void CircuitImpOpt::calcOptForks(const char* forks_name)
{
  if ( !top ) calcTop();
  if ( !top )
    Warning("cannot calc optimised isochronic forks because the top module is not defined");
  else
  {
    for (int module_index = 0; module_index < modules->count(); module_index++)
      (*modules)[module_index]->removeTag(TAG_PROCESSED);
    int forks_num = calcModuleOptForks(top, forks_name);
    for (int port_index = 0; port_index < top->ports->count(); port_index++)
    {
      Port* port_cur = (*top->ports)[port_index];
      if ( (port_cur->getType() == PORT_TYPE_OUTPUT)
        || (port_cur->getType() == PORT_TYPE_INOUT) )
        forks_num += port_cur->nets->count();
    }
    Implementation* implementation_forks = new Implementation(forks_name, IMPLEMENTATION_TYPE_FORKS, forks_num);
    addImplementation(implementation_forks);
  }
}

//---------------------------------------------------------------------------
int CircuitImpOpt::calcModuleOrphans(Module* module, const char* orphans_name)
{
  int result = 0;
  if (module->getType() == MODULE_TYPE_COMPOUND)
  {
    Property* property_cur = findProperty(module->getName());
    if (module->testTag(TAG_PROCESSED))
    {
      Implementation* implementation_orphans = property_cur->findImplementation(orphans_name);
      result = implementation_orphans->getValue();
    }
    else
    {
      // initialise
      for (int port_index = 0; port_index < module->ports->count(); port_index++)
      {
        Port* port_cur = (*module->ports)[port_index];
        if ( !port_cur->testTag(TAG_HIDDEN) 
          && ( (port_cur->getType() == PORT_TYPE_INPUT) || (port_cur->getType() == PORT_TYPE_WIRE) ) )
        {
          for (int net_index = 0; net_index < port_cur->nets->count(); net_index++)
          {
            Net* net_cur = (*port_cur->nets)[net_index];
            net_cur->addTag(TAG_ORPHAN);
          }
        }
      }
      // calculate
      for (int instance_index = 0; instance_index < module->instances->count(); instance_index++)
      {
        Instance* instance_cur = (*module->instances)[instance_index];
        Module* module_prot = findModule(instance_cur->getModuleName());
        if (module_prot->getType() == MODULE_TYPE_COMPOUND)
        {
          result += calcModuleOrphans(module_prot, orphans_name);
          for (int connection_index =  0; connection_index < instance_cur->connections->count(); connection_index++)
          {
            Connection* connection_cur = (*instance_cur->connections)[connection_index];
            if ( (connection_cur->getType() == CONNECTION_TYPE_INPUT)
              || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
            {
              for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
              {
                Net* net_cur = (*connection_cur->nets)[net_index];
                if (net_cur != NULL)
                  net_cur->removeTag(TAG_ORPHAN);
              }
            }
          }
        }
        else
        {
          int input_count = module_prot->getPortCount(PORT_TYPE_INPUT) + module_prot->getPortCount(PORT_TYPE_INOUT);
          int code_all = (1 << input_count) - 1;
          int code_cur = 0;
          Implementation* implementation_cur = (Implementation*)instance_cur->getData();
          if (implementation_cur->getType() == IMPLEMENTATION_TYPE_PARTIAL)
            code_cur = calcVariationCode(implementation_cur->getName(), module_prot);
          else
          if (implementation_cur->getType() == IMPLEMENTATION_TYPE_FULL)
            code_cur = code_all;
          int input_index = -1;
          for (int connection_index =  0; connection_index < instance_cur->connections->count(); connection_index++)
          {
            Connection* connection_cur = (*instance_cur->connections)[connection_index];
            if ( (connection_cur->getType() == CONNECTION_TYPE_INPUT)
              || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
            {
              input_index++;
              int input_code = (1 << input_index);
              if ((input_code & code_cur) == input_code)
              {
                for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
                {
                  Net* net_cur = (*connection_cur->nets)[net_index];
                  if (net_cur != NULL)
                    net_cur->removeTag(TAG_ORPHAN);
                }
              }
            }
          }
        }
      }      
      // count
      for (int port_index = 0; port_index < module->ports->count(); port_index++)
      {
        Port* port_cur = (*module->ports)[port_index];
        if ( !port_cur->testTag(TAG_HIDDEN) 
          && ( (port_cur->getType() == PORT_TYPE_INPUT) || (port_cur->getType() == PORT_TYPE_WIRE) ) )
        {
          for (int net_index = 0; net_index < port_cur->nets->count(); net_index++)
          {
            Net* net_cur = (*port_cur->nets)[net_index];
            if (net_cur->testTag(TAG_ORPHAN))
            {
              char* net_full_name = net_cur->genFullName();
              Warning("orphan net <%s> with %d potential indicators found in module <%s>", 
                net_full_name, net_cur->postset->count(), module->getName());
              delete [] net_full_name;
              if (net_cur->postset->count() > 0)
                result++;
            }
          }
        }
      }
      Implementation* implementation_orphans = new Implementation(orphans_name, IMPLEMENTATION_TYPE_ORPHANS, result);
      property_cur->addImplementation(implementation_orphans);
    }
  }
  return result;
}

//---------------------------------------------------------------------------
void CircuitImpOpt::calcOrphans(const char* orphans_name)
{
  if ( !top ) calcTop();
  if ( !top )
    Warning("cannot check indication because the top module is not defined");
  else
  {
    for (int module_index = 0; module_index < modules->count(); module_index++)
      (*modules)[module_index]->removeTag(TAG_PROCESSED);
    int orphans_num = calcModuleOrphans(top, orphans_name);
    Implementation* implementation_orphans = new Implementation(orphans_name, IMPLEMENTATION_TYPE_ORPHANS, orphans_num);
    addImplementation(implementation_orphans);
  }
}
