#include <stdio.h>
#include "sv_module.h"
#include "sv_object.h"
#include "sv_parameter.h"
#include "sv_port.h"
#include "sv_instance.h"
#include "messages.h"

//---------------------------------------------------------------------------
Module::Module(const char* _name, ModuleType _type, ParameterArray* _parameters, PortArray* _ports, InstanceArray* _instances) 
: Object(_name), type(_type), depth(-1), width(1)
{
  users = new InstanceArray(0);
  // create parameters	
  if (_parameters)
    parameters = new ParameterArray(_parameters->count());
  else
    parameters = new ParameterArray(0);
  addParameters(_parameters);
  // create ports
  port_hash_set = new PortHashSet();
  if (_ports)
    ports = new PortArray(_ports->count());
  else
    ports = new PortArray(5);
  addPorts(_ports);
  port_vdd = new Port(NET_NAME_VDD, PORT_TYPE_WIRE);
  port_vdd->addTag(TAG_HIDDEN);
  addPort(port_vdd);
  port_gnd = new Port(NET_NAME_GND, PORT_TYPE_WIRE);
  port_gnd->addTag(TAG_HIDDEN);
  addPort(port_gnd);
  // create instances
  if (_instances)
    instances = new InstanceArray(_instances->count());
  else
    instances = new InstanceArray(3);
  addInstances(_instances);
}

//---------------------------------------------------------------------------
Module::Module(const Module& module) 
: Object(module), type(module.type), depth(module.depth), width(module.width)
{
  // copy users
  users = new InstanceArray(module.users->count());
  addUsers(module.users);
  // copy parameters
  parameters = new ParameterArray(module.parameters->count());
  for (int parameter_index = 0; parameter_index < module.parameters->count(); parameter_index++)
    addParameter( new Parameter(*(*module.parameters)[parameter_index]) );
  // copy ports
  port_hash_set = new PortHashSet();
  ports = new PortArray(module.ports->count());
  for (int port_index = 0; port_index < module.ports->count(); port_index++)
    addPort( new Port(*(*module.ports)[port_index]) );
  port_vdd = findPort(module.port_vdd->getName());
  port_gnd = findPort(module.port_gnd->getName());
  // copy instances
  instances = new InstanceArray(module.instances->count());
  for (int instance_index = 0; instance_index < module.instances->count(); instance_index++)
    addInstance( new Instance(*(*module.instances)[instance_index]) );
  // recalculate connection nets to this module ports
  calcConnections();
}

//---------------------------------------------------------------------------
Module::~Module()
{
  clean();
  delete users; users = NULL;
  delete instances; instances = NULL;
  delete parameters; parameters = NULL;
  delete port_hash_set; port_hash_set = NULL;
  delete ports; ports = NULL;
}

//---------------------------------------------------------------------------
void Module::clean()
{
  instances->clean(true);
  parameters->clean(true);
  ports->clean(true);
}

//---------------------------------------------------------------------------
void Module::addUser(Instance* user, int position)
{
  if (position < 0)
    users->add(user);
  else
    users->insert(user, position);
}

//---------------------------------------------------------------------------
void Module::addUsers(InstanceArray* _users)
{
  if (_users)
  for (int user_index = 0; user_index < _users->count(); user_index++)
  {
    Instance* user_cur = (*_users)[user_index];
    addUser(user_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removeUser(Instance* user)
{
  users->remove(user, false);
}

//---------------------------------------------------------------------------
void Module::addParameter(Parameter* parameter, int position)
{
  if (position < 0)
    parameters->add(parameter);
  else
    parameters->insert(parameter, position);
}

//---------------------------------------------------------------------------
void Module::addParameters(ParameterArray* _parameters)
{
  if (_parameters)
  for (int parameter_index = 0; parameter_index < _parameters->count(); parameter_index++)
  {
    Parameter* parameter_cur = (*_parameters)[parameter_index];
    addParameter(parameter_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removeParameter(Parameter* parameter)
{
  parameters->remove(parameter, true);
}

//---------------------------------------------------------------------------
Parameter* Module::findParameter(const char* parameter_name) const
{
  Parameter* parameter_mask = new Parameter(parameter_name);
  int parameter_index = parameters->find(parameter_mask, 0);
  delete parameter_mask;
  if (parameter_index >= 0)
    return  (*parameters)[parameter_index];
  else
    return NULL;
}

//---------------------------------------------------------------------------
void Module::addPort(Port* port, int position)
{
  if (position < 0)
    ports->add(port);
  else
    ports->insert(port, position);
  port_hash_set->add(port);
}

//---------------------------------------------------------------------------
void Module::addPorts(PortArray* _ports)
{
  if (_ports)
  for (int port_index = 0; port_index < _ports->count(); port_index++)
  {
    Port* port_cur = (*_ports)[port_index];
    addPort(port_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removePort(Port* port)
{
  port_hash_set->remove(port);
  ports->remove(port, true);
}

//---------------------------------------------------------------------------
Port* Module::findPort(const char* port_name) const
{
  Port* port_mask = new Port(port_name);
  Port* port_cur = port_hash_set->find(port_mask);
  delete port_mask;
  return port_cur;
}

//---------------------------------------------------------------------------
int Module::getPortCount(PortType port_type) const
{
  int port_num = 0;
  for (int port_index = 0; port_index < ports->count(); port_index++)
  {
     Port* port_cur = (*ports)[port_index];
     if ( !port_cur->testTag(TAG_HIDDEN) )
       if (port_cur->getType() == port_type)
         port_num++;
  }
  return port_num;
}

//---------------------------------------------------------------------------
int Module::sortPorts(PortArray* _ports) 
{
  PortArray* ports_sorted = new PortArray(ports->count());
  // add the interface of the ports
  for (int port_index = 0; port_index < _ports->count(); port_index++)
  {
    Port* port_new = (*_ports)[port_index];
    Port* port_cur = findPort(port_new->getName());
    if (port_cur)
    {
      ports_sorted->add(port_cur);
      port_cur->addTag(TAG_PROCESSED);
    }
    else
      Error("port <%s> of module <%s> is not defined", port_new->getName(), name);
  }
  // add the rest of the ports
  for (int port_index = 0; port_index < ports->count(); port_index++)
  {
    Port* port_cur = (*ports)[port_index];
    if (port_cur->testTag(TAG_PROCESSED))
      port_cur->removeTag(TAG_PROCESSED);
    else
      ports_sorted->add(port_cur);
  }
  // change the port list in the module
  delete ports;
  ports = ports_sorted;
}

//---------------------------------------------------------------------------
Net* Module::findNet(const char* net_name, int net_index) const
{
  Port* port_cur = findPort(net_name);
  if (port_cur)
    return port_cur->findNet(net_index);
  else
    return NULL;
}

//---------------------------------------------------------------------------
void Module::addInstance(Instance* instance, int position)
{
  instance->setContainer(this);
  if (position < 0)
    instances->add(instance);
  else
    instances->insert(instance, position);
}

//---------------------------------------------------------------------------
void Module::addInstances(InstanceArray* _instances)
{
  if (_instances)
  for (int instance_index = 0; instance_index < _instances->count(); instance_index++)
  {
    Instance* instance_cur = (*_instances)[instance_index];
    addInstance(instance_cur);
  }
}

//---------------------------------------------------------------------------
void Module::removeInstance(Instance* instance)
{
  instances->remove(instance, true);
}

//---------------------------------------------------------------------------
Instance* Module::findInstance(const char* instance_name, int instance_index) const
{
  Instance* instance_mask = new Instance(instance_name);
  int instance_find_index = instances->find(instance_mask, 0);
  delete instance_mask;
  if (instance_find_index >= 0)
  {
    Instance* instance = (*instances)[instance_find_index];
    if (instance_index == instance->getIndex())
      return instance;
  }
  return NULL;
}

//---------------------------------------------------------------------------
void Module::calcConnections()
{
  NetArray* nets_tmp = new NetArray(1);
  for (int instance_index = 0; instance_index < instances->count(); instance_index++)
  {
    Instance* instance_cur = (*instances)[instance_index];
    char* instance_full_name = instance_cur->genFullName();
    for (int connection_index = 0; connection_index < instance_cur->connections->count(); connection_index++)
    {
      Connection* connection_cur = (*instance_cur->connections)[connection_index];
      for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
      {
        Net* net_cur = (*connection_cur->nets)[net_index];
        if (net_cur)
        {
          bool net_undefined = false;
          Net* net_new = NULL;
          Port* port_cur = findPort(net_cur->getName());
          if ( !port_cur )
          {
            if (net_cur->getIndex() == -1)
            {
//              Warning("in module <%s> instance <%s> is connected to undeclared net <%s>", 
//                name, instance_full_name, net_cur->getName());
              port_cur = new Port(net_cur->getName(), PORT_TYPE_WIRE);
              addPort(port_cur);
            }
          }
          if (port_cur)
          {
            // connection of a bus
            if (net_cur->getIndex() == -1)
            {
              for (int net_new_index = 0; net_new_index < port_cur->nets->count(); net_new_index++)
              {
                net_new = (*port_cur->nets)[net_new_index];
                if (net_new != net_cur)
                {
                  if (net_new_index == 0)
                    (*connection_cur->nets)[net_index] = net_new;
                  else
                    connection_cur->nets->insert(net_new, ++net_index);
                }
              }
            }
            // connection of a wire
            else
            {
              net_new = port_cur->findNet(net_cur->getIndex());
              if (net_new != net_cur)
              {
                if (net_new)
                  (*connection_cur->nets)[net_index] = net_new;
                else
                  net_undefined = true;
              }
            }
          }
          else
            net_undefined = true;
          
          // remember the nets which can be removed in the end  
          if (net_cur != net_new)
            if ( nets_tmp->locate(net_cur) < 0 )
               nets_tmp->add(net_cur);
 
          if (net_undefined)
          {
            char* net_full_name = net_cur->genFullName();
            Error("in module <%s> instance <%s> is connected to undeclared net <%s> which is in a bus",
              name, instance_full_name, net_full_name);
            delete [] net_full_name;
          }
        }
      }
    }
    delete [] instance_full_name;
  }
  // clean unused nets
  nets_tmp->clean(true);
  delete nets_tmp; nets_tmp = NULL;
}

//---------------------------------------------------------------------------
ConnectionType Module::castConnectionType(PortType port_type)
{
  ConnectionType connection_type = CONNECTION_TYPE_UNDEFINED;
  switch (port_type)
  {
    case PORT_TYPE_INPUT:
      connection_type = CONNECTION_TYPE_INPUT;
      break;
    case PORT_TYPE_OUTPUT:
      connection_type = CONNECTION_TYPE_OUTPUT;
      break;
    case PORT_TYPE_INOUT:
      connection_type = CONNECTION_TYPE_INOUT;
      break;
  }
  return connection_type;
}

//---------------------------------------------------------------------------
void Module::sortConnections()
{
  for (int instance_index = 0; instance_index < instances->count(); instance_index++)
  {
    Instance* instance_cur = (*instances)[instance_index];
    char* instance_full_name = instance_cur->genFullName();
    Module* module_prot = container->findModule(instance_cur->getModuleName());
    // if prototype is found
    if (module_prot)
    {
      for (int connection_index = 0; connection_index < instance_cur->connections->count(); connection_index++)
      {
        Connection* connection_cur = (*instance_cur->connections)[connection_index];
        if (connection_cur->getName() == NULL)
        {
          if (module_prot->ports)
          {
            int port_index = -1;
            int port_connection_index = port_index;
            while (port_connection_index < connection_index)
            {
              port_index++;
              if (port_index < module_prot->ports->count())
              {
                Port* port_cur = (*module_prot->ports)[port_index];
                if ( !port_cur->testTag(TAG_HIDDEN) )
                  if ( (port_cur->getType() == PORT_TYPE_INPUT)
                    || (port_cur->getType() == PORT_TYPE_OUTPUT) 
                    || (port_cur->getType() == PORT_TYPE_INOUT) )
                    port_connection_index++;
              }
              else
                Error("instance <%s> of module <%s> in module <%s> has more connections then ports available",
                  instance_full_name, instance_cur->getModuleName(), name);
            }
            if (port_connection_index == connection_index)
            {
              Port* port_cur = (*module_prot->ports)[port_index];
              connection_cur->setType( castConnectionType(port_cur->getType()) );
              connection_cur->setName( port_cur->getName() );
            }
          }
        }
      }
      // sort connections according to port order
      if (module_prot->ports)
      {
        int connection_index_max = 0;
        for (int port_index = 0; port_index < module_prot->ports->count(); port_index++)
        {
          Port* port_cur = (*module_prot->ports)[port_index];
          if ( !port_cur->testTag(TAG_HIDDEN) )
          {
            if ( (port_cur->getType() == PORT_TYPE_INPUT) 
              || (port_cur->getType() == PORT_TYPE_OUTPUT)
              || (port_cur->getType() == PORT_TYPE_INOUT) )
            {
              bool connection_found = false;
              for (int connection_index = connection_index_max; (!connection_found) && (connection_index < instance_cur->connections->count()); connection_index++)
              {
                Connection* connection_cur = (*instance_cur->connections)[connection_index];
                if (connection_cur->getName())
                if ( strcmp(port_cur->getName(), connection_cur->getName()) == 0 )
                {
                  connection_found = true;
                  (*instance_cur->connections)[connection_index] = (*instance_cur->connections)[connection_index_max];
                  (*instance_cur->connections)[connection_index_max] = connection_cur;
                  connection_cur->setType( castConnectionType(port_cur->getType()) );
                  connection_index_max++;
                }
              }
              if (!connection_found)
              {
                Connection* connection_new = new Connection(port_cur->getName(), castConnectionType(port_cur->getType()) );
                instance_cur->addConnection(connection_new, connection_index_max);
                connection_index_max++;
              }
            }
          }
        }
      }
    }
    else
      Error("instance <%s> of module <%s> in module <%s> has no prototype!!!", 
        instance_full_name, instance_cur->getModuleName(), name);
    // check if all connection are associated to ports 
    for (int connection_index = 0; connection_index < instance_cur->connections->count(); connection_index++)
    {
      Connection* connection_cur = (*instance_cur->connections)[connection_index];
      if (connection_cur->getName() == NULL)
        Error("some ports of instance <%s> of module <%s> in module <%s> cannot be associated", 
          instance_full_name, instance_cur->getModuleName(), name);
    }
    delete [] instance_full_name;
  }
}

//---------------------------------------------------------------------------
void Module::calcIncidence()
{
  // clean all nets preset and postset
  for (int port_index = 0; port_index < ports->count(); port_index++)
  {
    Port* port_cur = (*ports)[port_index];
    for (int net_index = 0; net_index < port_cur->nets->count(); net_index++)
    {
      Net* net_cur = (*port_cur->nets)[net_index];
      net_cur->cleanPreset();
      net_cur->cleanPostset();
    }
  }
  // calc new preset and postset
  for (int instance_index = 0; instance_index < instances->count(); instance_index++)
  {
    Instance* instance_cur = (*instances)[instance_index];
    if (instance_cur->connections)
    for (int connection_index = 0; connection_index < instance_cur->connections->count(); connection_index++)
    {
      Connection* connection_cur = (*instance_cur->connections)[connection_index];
      for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
      {
        Net* net_cur = (*connection_cur->nets)[net_index];
        if (net_cur)
        {
          if ( (connection_cur->getType() == CONNECTION_TYPE_INPUT)
            || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
            net_cur->addPostset(instance_cur);
          if ( (connection_cur->getType() == CONNECTION_TYPE_OUTPUT)
            || (connection_cur->getType() == CONNECTION_TYPE_INOUT) )
            net_cur->addPreset(instance_cur);
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void Module::findHierarchyNet(Net* net, int depth, int layer)
{
  if (net && !net->testTag(TAG_PROCESSED))
  {
  	int depth_cur = net->getDepth();
  	int layer_cur = net->getLayer();
    if ( (depth > depth_cur) || (layer < layer_cur) || (layer_cur < 0) )
    {
      net->addTag(TAG_PROCESSED);
      if (depth > depth_cur)
        net->setDepth(depth);
      if ((layer < layer_cur) || (layer_cur < 0))
        net->setLayer(layer);
      if (net->preset)
      {
        if (net->preset->count() > 1)
        {
          char* net_name = net->genFullName();
          Error("net <%s> in module <%s> is an output of more then 1 instance", net_name, name);
          delete [] net_name;
        }
        else
        if (net->preset->count() == 1)
        {
          Instance* instance_cur = (*net->preset)[0];
          findHierarchyInstance(instance_cur, depth, layer);
        }
      }
      net->removeTag(TAG_PROCESSED);
    }
  }
}

//---------------------------------------------------------------------------
void Module::findHierarchyInstance(Instance* instance, int depth, int layer)
{
  if (instance && !instance->testTag(TAG_PROCESSED))
  {
  	int depth_cur = instance->getDepth();
  	int layer_cur = instance->getLayer();
    int instance_width = -1;
    Module* module_prot = container->findModule(instance->getModuleName());
    if (module_prot)
      instance_width = module_prot->getWidth();
    if (instance_width < 0)
      instance_width = 1;

    if ((depth > depth_cur) || (layer < layer_cur) || (layer_cur < 0))
    {
      instance->addTag(TAG_PROCESSED);
      if (depth > depth_cur)
        instance->setDepth(depth);
      if ((layer < layer_cur) || (layer_cur < 0))
        instance->setLayer(layer);
    
      depth = instance->getDepth() + instance_width;
      layer = instance->getLayer() + instance_width;
      if (instance->connections)
      for (int connection_index = 0; connection_index < instance->connections->count(); connection_index++)
      {
        Connection* connection_cur = (*instance->connections)[connection_index];
        if (connection_cur->getType() == CONNECTION_TYPE_INPUT)
        {
          for (int net_index = 0; net_index < connection_cur->nets->count(); net_index++)
          {
            Net* net_cur = (*connection_cur->nets)[net_index];
            findHierarchyNet(net_cur, depth, layer);
          }
        }
      }
      instance->removeTag(TAG_PROCESSED);
    }
  }
}

//---------------------------------------------------------------------------
void Module::calcHierarchy()
{
  calcIncidence();
  cleanHierarchy();
  // calc hierarchy (going from outputs)
  for (int port_index = 0; port_index < ports->count(); port_index++)
  {
    Port* port_cur = (*ports)[port_index];
    if (port_cur->getType() == PORT_TYPE_OUTPUT)
    {
      for (int net_index = 0; net_index < port_cur->nets->count(); net_index++)
      {
        Net* net_cur = (*port_cur->nets)[net_index];
        findHierarchyNet(net_cur, 0, 0);
      }
    }
  }
  calcWidth();
  checkHierarchy();
}


//---------------------------------------------------------------------------
void Module::cleanHierarchy()
{
  // initialise instances
  for (int instance_index = 0; instance_index < instances->count(); instance_index++)
  {
    Instance* instance_cur = (*instances)[instance_index];
    instance_cur->removeTag(TAG_PROCESSED);
    instance_cur->setDepth(-1);
    instance_cur->setLayer(-1);
  }
  // initialise nets
  for (int port_index = 0; port_index < ports->count(); port_index++)
  {
    Port* port_cur = (*ports)[port_index];
    for (int net_index = 0; net_index < port_cur->nets->count(); net_index++)
    {
      Net* net_cur = (*port_cur->nets)[net_index];
      net_cur->removeTag(TAG_PROCESSED);
      net_cur->setDepth(-1);
      net_cur->setLayer(-1);
    }
  }
}

//---------------------------------------------------------------------------
void Module::checkHierarchy()
{
  // show unused instances
  for (int instance_index = 0; instance_index < instances->count(); instance_index++)
  {
    Instance* instance_cur = (*instances)[instance_index];
    if (instance_cur->getDepth() < 0)
    if ( !instance_cur->testTag(TAG_HIDDEN) )
    {
      char* instance_name = instance_cur->genFullName();
      Warning("instance <%s> of module <%s> in module <%s> does not influence the output", 
        instance_name, instance_cur->getModuleName(), name);
      delete [] instance_name;
    }
  }
  // show unused nets
  for (int port_index = 0; port_index < ports->count(); port_index++)
  {
    Port* port_cur = (*ports)[port_index];
    for (int net_index = 0; net_index < port_cur->nets->count(); net_index++)
    {
      Net* net_cur = (*port_cur->nets)[net_index];
      if (net_cur->getDepth() < 0)
      if ( !net_cur->getContainer()->testTag(TAG_HIDDEN) )
      {
        char* net_name = net_cur->genFullName();
        Warning("net <%s> in module <%s> does not influence the output", net_name, name);
        delete [] net_name;
      }
    }
  }
}

//---------------------------------------------------------------------------
void Module::calcWidth()
{
  // calc width of the module
  width = -1;
  for (int instance_index = 0; instance_index < instances->count(); instance_index++)
  {
    Instance* instance_cur = (*instances)[instance_index];
    int instance_depth = instance_cur->getDepth();
    int instance_width = -1;
    Module* module_prot = container->findModule(instance_cur->getModuleName());
    if (module_prot)
      instance_width = module_prot->getWidth();
    if (instance_width < 0)
      instance_width = 1;
    if ( (instance_depth >= 0) && (instance_width >= 0) )
    {
      int width_cur = instance_depth + instance_width;
      if (width_cur > width)
        width = width_cur;
    }
  }
}
