#ifndef _SV_MODULE_H_
#define _SV_MODULE_H_

#include "array.h"
#include "hashset.h"
#include "sv_object.h"

class Module;
typedef Array <Module, ObjectCompareName> ModuleArray;
typedef HashSet <Module, ObjectHashName, ObjectEqualName> ModuleHashSet;

enum ModuleType
{
  MODULE_TYPE_UNDEFINED,
  MODULE_TYPE_ELEMENTARY,
  MODULE_TYPE_COMPOUND
};

#include "sv_parameter.h"
#include "sv_net.h"
#include "sv_port.h"
#include "sv_connection.h"
#include "sv_instance.h"
#include "sv_circuit.h"

//---------------------------------------------------------------------------
class Module : public Object
{
  public:
    ParameterArray* parameters;
    PortArray* ports;
    InstanceArray* instances;
    InstanceArray* users;

    Module(const char* _name, ModuleType _type = MODULE_TYPE_UNDEFINED, ParameterArray* _parameters = NULL, 
      PortArray* _ports = NULL, InstanceArray* _instances = NULL);
    Module(const Module& module);
    ~Module();
    
    void setType(ModuleType _type);
    ModuleType getType() const;
    void setDepth(int _depth);
    int getDepth() const;
    void setWidth(int _width);
    int getWidth() const;
    void setContainer(Circuit* container);
    Circuit* getContainer() const;

    void addUser(Instance* user, int position = -1);
    void addUsers(InstanceArray* _users);
    void removeUser(Instance* user);

    void addParameter(Parameter* parameter, int position = -1);
    void addParameters(ParameterArray* _parameters);
    void removeParameter(Parameter* parameter);
    Parameter* findParameter(const char* parameter_name) const;

    void addPort(Port* port, int position = -1);
    void addPorts(PortArray* _ports);
    void removePort(Port* port);
    Port* findPort(const char* port_name) const;
    int getPortCount(PortType port_type) const;
    int sortPorts(PortArray* _ports);
    Net* findNet(const char* net_name, int net_index) const;

    void addInstance(Instance* instance, int position = -1);
    void addInstances(InstanceArray* _instances);
    void removeInstance(Instance* instance);
    Instance* findInstance(const char* instance_name, int instance_index = -1) const;

    void calcConnections();
    void sortConnections();
    void calcHierarchy();

  protected:
    ModuleType type;
    int depth;
    int width;
    Port* port_vdd;
    Port* port_gnd;
    Circuit* container;
    PortHashSet* port_hash_set;

    void clean();
    ConnectionType castConnectionType(PortType port_type);
    void calcIncidence();
    void findHierarchyNet(Net* net, int depth, int layer);
    void findHierarchyInstance(Instance* instance, int depth, int layer);
    void cleanHierarchy();
    void checkHierarchy();
    void calcWidth();
};

//---------------------------------------------------------------------------
inline void Module::setType(ModuleType _type)
{
  type = _type;
}

//---------------------------------------------------------------------------
inline ModuleType Module::getType() const
{
  return type;
}
//---------------------------------------------------------------------------
inline void Module::setDepth(int _depth)
{
  depth = _depth;
}

//---------------------------------------------------------------------------
inline int Module::getDepth() const
{
  return depth;
}

//---------------------------------------------------------------------------
inline void Module::setWidth(int _width)
{
  width = _width;
}

//---------------------------------------------------------------------------
inline int Module::getWidth() const
{
  return width;
}

//---------------------------------------------------------------------------
inline void Module::setContainer(Circuit* _container)
{
  container = _container;
}

//---------------------------------------------------------------------------
inline Circuit* Module::getContainer() const
{
  return container;
}

#endif
