%{
#include "messages.h"
#include "str_utils.h"
#include "sv_object.h"
#include "sv_delay.h"
#include "sv_range.h"
#include "sv_parameter.h"
#include "sv_connection.h"
#include "sv_net.h"
#include "sv_port.h"
#include "sv_instance.h"
#include "sv_module.h"
#include "sv_circuit.h"
#include "sverilog.h"

//#define DEBUG_YACC 1
#ifdef DEBUG_YACC
  #include <stdio.h>
  #define DEBUG_YACC_MESSAGE(msg) fprintf(stdout, msg)
#else
  #define DEBUG_YACC_MESSAGE(msg) // msg
#endif

extern int sverilog_lex();

void sverilog_error(Circuit* circuit, const char* file_name, ModuleType module_type, char* message)
{
  Error("%s in line %d in file <%s>", message, sverilog_lineno, file_name);
}

%}

%parse-param {Circuit* circuit}
%parse-param {const char* file_name}
%parse-param {ModuleType module_type}
%start source_text

%union
{
  char* string;
  int inumber;
  float rnumber;
  Range* range;
  Delay* delay;
  ObjectArray* objects;
  Parameter* parameter;
  ParameterArray* parameters;
  Port* port;
  PortArray* ports;
  Instance* instance;
  InstanceArray* instances;
  Connection* connection;
  ConnectionArray* connections;
  Module* module;
  ModuleArray* modules;
}

%token <string> YYMODULE
%token <string> YYENDMODULE
%token <string> YYPARAMETER
%token <string> YYINPUT
%token <string> YYOUTPUT
%token <string> YYINOUT
%token <string> YYWIRE
%token <string> YYREG
%token <string> YYID
%token <inumber> YYINUMBER
%token <rnumber> YYRNUMBER

%type <module> module
%type <ports> port_list_opt
%type <ports> port_list                     // creates PortArray
%type <port> port
%type <port> port_expression_opt
%type <port> port_expression
%type <port> port_reference                 // creates Port
%type <range> port_reference_arg            // creates Range
%type <module> module_item_clr              // creates Module
%type <parameters> parameter_declaration    // creates PortArray
%type <ports> port_declaration              // creates PortArray
%type <ports> input_declaration
%type <ports> inout_declaration
%type <ports> output_declaration
%type <ports> reg_declaration
%type <ports> wire_declaration
%type <instances> instance_declaration
%type <instances> instance_list             // creates InstanceArray
%type <instance> instance                   // creates Instance
%type <string> instance_name_opt
%type <inumber> instance_index_opt
%type <connections> connection_list
%type <connections> ordered_connection_list // creates ConnectionArray
%type <connections> named_connection_list   // creates ConnectionArray
%type <connection> named_connection
%type <connection> connection               // creates Connection
%type <objects> variable_list               // creates ObjectArray
%type <range> range_opt
%type <range> range                         // creates Range
%type <delay> delay_opt
%type <delay> delay                         // creates Delay
%type <parameters> assignment_list          // creates ParameterArray
%type <parameter> assignment                // creates Parameter
%type <string> lvalue
%type <rnumber> constant

%%

source_text
  :
    {
      DEBUG_YACC_MESSAGE("source_text : \n");
    }
  | source_text module
    {
      DEBUG_YACC_MESSAGE("source_text : source_text module\n");
      circuit->addModule($2);
    }
;

module
  : YYMODULE YYID port_list_opt ';' module_item_clr YYENDMODULE
    {
      DEBUG_YACC_MESSAGE("module : YYMODULE YYID port_list_opt ';' module_item_clr YYENDMODULE\n");
      $$ = $5;
      $$->setName($2);
      $$->setType(module_type);
      $$->sortPorts($3);
      $3->clean(true);
      delete [] $2;
      delete $3;
    }
;

module_item_clr
  :
    {
      DEBUG_YACC_MESSAGE("module_item_clr : \n");
      $$ = new Module(NULL);
    }
  | module_item_clr parameter_declaration
    {
      DEBUG_YACC_MESSAGE("module_item_clr : module_item_clr parameter_declaration\n");
      $$ = $1;
      for (int i = 0; i < $2->count(); i++)
        $$->addParameter( (*$2)[i] );
      delete $2;
    }
  | module_item_clr port_declaration
    {
      DEBUG_YACC_MESSAGE("module_item_clr : module_item_clr port_declaration\n");
      $$ = $1;
      for (int i = 0; i < $2->count(); i++)
        $$->addPort( (*$2)[i] );
      delete $2;
    }
  | module_item_clr instance_declaration
    {
      DEBUG_YACC_MESSAGE("module_item_clr : module_item_clr instance_declaration\n");
      $$ = $1;
      for (int i = 0; i < $2->count(); i++)
        $$->addInstance( (*$2)[i] );
      delete $2;
    }
;

parameter_declaration
  : YYPARAMETER assignment_list ';'
    {
      DEBUG_YACC_MESSAGE("parameter_declaration : YYPARAMETER assignment_list ';'\n");
      $$ = $2;
    }
;

port_declaration
  : input_declaration
    {
      DEBUG_YACC_MESSAGE("module_item : input_declaration\n");
      $$ = $1;
    }
  | output_declaration
    {
      DEBUG_YACC_MESSAGE("module_item : output_declaration\n");
      $$ = $1;
    }
  | inout_declaration
    {
      DEBUG_YACC_MESSAGE("module_item : inout_declaration\n");
      $$ = $1;
    }
  | reg_declaration
    {
      DEBUG_YACC_MESSAGE("module_item : reg_declaration\n");
      $$ = $1;
    }
  | wire_declaration
    {
      DEBUG_YACC_MESSAGE("module_item : wire_declaration\n");
      $$ = $1;
    }
;

input_declaration
  : YYINPUT range_opt variable_list ';'
    {
      DEBUG_YACC_MESSAGE("input_declaration : YYINPUT range_opt variable_list ';'\n");
      $$ = new PortArray($3->count());
      for (int i = 0; i < $3->count(); i++)
      {
        Object* object_cur = (*$3)[i];
        Port* port_new = new Port(object_cur->getName(), PORT_TYPE_INPUT, $2);
        $$->add(port_new);
      }
      delete $2;
      $3->clean(true);
      delete $3;
    }
;

output_declaration
  : YYOUTPUT range_opt variable_list ';'
    {
      DEBUG_YACC_MESSAGE("output_declaration : YYOUTPUT range_opt variable_list ';'\n");
      $$ = new PortArray($3->count());
      for (int i = 0; i < $3->count(); i++)
      {
        Object* object_cur = (*$3)[i];
        Port* port_new = new Port(object_cur->getName(), PORT_TYPE_OUTPUT, $2);
        $$->add(port_new);
      }
      delete $2;
      $3->clean(true);
      delete $3;
    }
;

inout_declaration
  : YYINOUT range_opt variable_list ';'
    {
      DEBUG_YACC_MESSAGE("inout_declaration : YYINOUT range_opt variable_list ';'\n");
      $$ = new PortArray($3->count());
      for (int i = 0; i < $3->count(); i++)
      {
        Object* object_cur = (*$3)[i];
        Port* port_new = new Port(object_cur->getName(), PORT_TYPE_INOUT, $2);
        $$->add(port_new);
      }
      delete $2;
      $3->clean(true);
      delete $3;
    }
;

reg_declaration
  : YYREG range_opt variable_list ';'
    {
      DEBUG_YACC_MESSAGE("reg_declaration : YYREG range_opt variable_list ';'\n");
      $$ = new PortArray($3->count());
      for (int i = 0; i < $3->count(); i++)
      {
        Object* object_cur = (*$3)[i];
        Port* port_new = new Port(object_cur->getName(), PORT_TYPE_REG, $2);
        $$->add(port_new);
      }
      delete $2;
      $3->clean(true);
      delete $3;
    }
;

wire_declaration
  : YYWIRE range_opt delay_opt variable_list ';'
    {
      DEBUG_YACC_MESSAGE("net_declaration : YYWIRE range_opt delay_opt variable_list ';'\n");
      $$ = new PortArray($4->count());
      for (int i = 0; i < $4->count(); i++)
      {
        Object* object_cur = (*$4)[i];
        // TODO: parse port dimension
        Port* port_new = new Port(object_cur->getName(), PORT_TYPE_WIRE, $2, $3);
        $$->add(port_new);
      }
      delete $2;
      delete $3;
      $4->clean(true);
      delete $4;
    }
;

instance_declaration
  : YYID instance_list ';'
  {
      $$ = $2;
      for (int i = 0; i < $$->count(); i++)
        (*$$)[i]->setModuleName($1);
      delete [] $1;
  }

instance_list
  : instance
    {
      DEBUG_YACC_MESSAGE("instance_list : instance\n");
      $$ = new InstanceArray(1);
      $$->add($1);
    }
  | instance_list ',' instance
    {
      DEBUG_YACC_MESSAGE("instance_list : instance_list ',' instance\n");
      $$ = $1;
      $$->add($3);
    }
;

instance
  : instance_name_opt instance_index_opt '(' connection_list ')'
    {
      DEBUG_YACC_MESSAGE("instance : instance_name_opt instance_index_opt '(' connection_list ')'\n");
      $$ = new Instance($1, $2, NULL, $4);
      delete [] $1;
      delete $4;
    }
;

instance_name_opt
  :
    {
      DEBUG_YACC_MESSAGE("instance_name_opt : \n");
      $$ = NULL;
    }
  | YYID
    {
      DEBUG_YACC_MESSAGE("instance_name_opt : YYID\n");
      $$ = $1;
    }
;

instance_index_opt
  :
    {
      DEBUG_YACC_MESSAGE("instance_index_opt : \n");
      $$ = -1;
    }
  | '[' YYINUMBER ']'
    {
      DEBUG_YACC_MESSAGE("instance_index_opt : '[' YYINUMBER ']'\n");
      $$ = $2;
    }

connection_list
  : ordered_connection_list
    {
      DEBUG_YACC_MESSAGE("connection_list : ordered_connection_list\n");
      $$ = $1;
    }
  | named_connection_list
    {
      DEBUG_YACC_MESSAGE("connection_list : named_connection_list\n");
      $$ = $1;
    }
;

ordered_connection_list
  : connection
    {
      DEBUG_YACC_MESSAGE("ordered_connection_list : connection\n");
      $$ = new ConnectionArray(1);
      $$->add($1);
    }
  | ordered_connection_list ',' connection
    {
      DEBUG_YACC_MESSAGE("ordered_connection_list : ordered_connection_list ',' connection\n");
      $$ = $1;
      $$->add($3);
    }
;

named_connection_list
  : named_connection
    {
      DEBUG_YACC_MESSAGE("named_connection_list : named_connection\n");
      $$ = new ConnectionArray(1);
      $$->add($1);
    }
  | named_connection_list ',' named_connection
    {
      DEBUG_YACC_MESSAGE("named_connection_list : named_connection_list ',' named_connection\n");
      $$ = $1;
      $$->add($3);
    }
;

named_connection
  : '.' YYID '(' connection ')'
    {
      DEBUG_YACC_MESSAGE("named_connection : '.' YYID '(' connection ')'\n");
      $$ = $4;
      $$->setName($2);
      delete [] $2;
    }
;

connection
  : port_expression_opt
    {
      DEBUG_YACC_MESSAGE("connection : port_expression_opt\n");
      NetArray* nets = NULL;
      if ($1)
      {
        nets = new NetArray($1->nets->count());
        for (int net_ind = 0; net_ind < $1->nets->count(); net_ind++)
        {
          Net* net_cur = (*$1->nets)[net_ind];
          if (net_cur)
          {
            Net* net_new = new Net(*net_cur);
            nets->add(net_new);
          }
          else
            nets->add(NULL);
        }
        delete $1;
      }
      $$ = new Connection(NULL, CONNECTION_TYPE_UNDEFINED, nets);
      delete nets;
    }
  | '{' port_list '}'
    {
      DEBUG_YACC_MESSAGE("connection : '{' port_list_opt '}'\n");
      NetArray* nets = NULL;
      if ($2)
      {
        nets = new NetArray($2->count());
        for (int port_ind = 0; port_ind < $2->count(); port_ind++)
        {
          Port* port_cur = (*$2)[port_ind];
          for (int net_ind = 0; net_ind < port_cur->nets->count(); net_ind++)
          {
            Net* net_cur = (*port_cur->nets)[net_ind];
            if (net_cur)
            {
              Net* net_new = new Net(*net_cur);
              nets->add(net_new);
            }
            else
              nets->add(NULL);
          }
          delete port_cur;
        }
        delete $2;
      }
      $$ = new Connection(NULL, CONNECTION_TYPE_UNDEFINED, nets);
      delete nets;
    }
;

port_list_opt
  :
    {
      DEBUG_YACC_MESSAGE("port_list_opt : \n");
      $$ = NULL;
    }
  | '(' port_list ')'
    {
      DEBUG_YACC_MESSAGE("port_list_opt : '(' port_list ')'\n");
      $$ = $2;
    }
;

port_list
  : port
    {
      DEBUG_YACC_MESSAGE("port_list : port\n");
      $$ = new PortArray(2);
      $$->add($1);
    }
  | port_list ',' port
    {
      DEBUG_YACC_MESSAGE("port_list : port_list ',' port\n");
      $$ = $1;
      $$->add($3);
    }
;

port
  : port_expression_opt
    {
      DEBUG_YACC_MESSAGE("port : port_expression_opt\n");
      $$ = $1;
    }
;

port_expression_opt
  :
    {
      DEBUG_YACC_MESSAGE("port_expression_opt : \n");
      $$ = NULL;
    }
  | port_expression
    {
      DEBUG_YACC_MESSAGE("port_expression_opt : port_expression\n");
      $$ = $1;
    }
;

port_expression
  : port_reference
    {
      DEBUG_YACC_MESSAGE("port_expression : port_reference\n");
      $$ = $1;
    }
;

port_reference
  : YYID port_reference_arg
    {
      DEBUG_YACC_MESSAGE("port_reference : YYID port_reference_arg\n");
      $$ = new Port($1, PORT_TYPE_UNDEFINED, $2);
      delete $2;
      delete [] $1;
    }
  | YYINUMBER
    {
      DEBUG_YACC_MESSAGE("port_reference : YYINUMBER\n");
      switch ($1)
      {
        case 0:
          $$ = new Port(NET_NAME_GND);
          break;
        case 1:
          $$ = new Port(NET_NAME_VDD);
          break;
      }
    }
;

port_reference_arg
  :
    {
      DEBUG_YACC_MESSAGE("port_reference_arg : \n");
      $$ = NULL;
    }
  | '[' YYINUMBER ']'
    {
      DEBUG_YACC_MESSAGE("port_reference_arg : '[' YYINUMBER ']'\n");
      $$ = new Range($2, $2);
    }
  | '[' YYINUMBER ':' YYINUMBER ']'
    {
      DEBUG_YACC_MESSAGE("port_reference_arg : '[' YYINUMBER ':' YYINUMBER ']'\n");
      $$ = new Range($2, $4);
    }
;

variable_list
  : YYID
    {
      DEBUG_YACC_MESSAGE("variable_list : YYID\n");
      $$ = new ObjectArray(1);
      Object* object_new = new Object($1);
      $$->add(object_new);
      delete [] $1;
    }
  | variable_list ',' YYID
    {
      DEBUG_YACC_MESSAGE("variable_list : variable_list ',' YYID\n");
      $$ = $1;
      Object* object_new = new Object($3);
      $$->add(object_new);
      delete [] $3;
    }
 ;

range_opt
  :
    {
      DEBUG_YACC_MESSAGE("range_opt : \n");
      $$ = NULL;
    }
  | range
    {
      DEBUG_YACC_MESSAGE("range_opt : range\n");
      $$ = $1;
    }
;

range
  : '[' YYINUMBER ':' YYINUMBER ']'
    {
      DEBUG_YACC_MESSAGE("range : '[' YYINUMBER ':' YYINUMBER ']'\n");
      $$ = new Range($2, $4);
    }
;

assignment_list
  : assignment
    {
      DEBUG_YACC_MESSAGE("assignment_list : assignment\n");
      $$ = new ParameterArray(1);
      $$->add($1);
    }
  | assignment_list ',' assignment
    {
      DEBUG_YACC_MESSAGE("assignment_list : assignment_list ',' assignment\n");
      $$ = $1;
      $$->add($3);
    }
;

assignment
  : lvalue '=' constant
    {
      DEBUG_YACC_MESSAGE("assignment : lvalue '=' constant\n");
      $$ = new Parameter($1, $3);
      delete [] $1;
    }
;

lvalue
  : YYID
    {
      DEBUG_YACC_MESSAGE("lvalue : YYID\n");
      $$= $1;
    }
;

delay_opt
  :
    {
      DEBUG_YACC_MESSAGE("delay_opt : \n");
      $$ = NULL;
    }
  | delay
    {
      DEBUG_YACC_MESSAGE("delay_opt : delay\n");
      $$ = $1;
  }
;

delay
  : '#' constant
    {
      DEBUG_YACC_MESSAGE("delay : '#' constant\n");
      $$ = new Delay($2, $2);
    }
  | '#' '(' constant ')'
    {
      DEBUG_YACC_MESSAGE("delay : '#' '(' constant ')'\n");
      $$ = new Delay($3, $3);
    }
  | '#' '(' constant ',' constant ')'
    {
      DEBUG_YACC_MESSAGE("delay : '#' '(' constant ',' constant ')'\n");
      $$ = new Delay($3, $5);
    }
;

constant
  : YYINUMBER
    {
      DEBUG_YACC_MESSAGE("constant : YYINUMBER\n");
      $$ = $1;
    }
  | YYRNUMBER
    {
      DEBUG_YACC_MESSAGE("constant : YYRNUMBER\n");
      $$ = $1;
    }
;

%%
